'''
CRKSPH corrections

These are equations for the basic kernel corrections in [CRKSPH2017].

References
-----------

    .. [CRKSPH2017] Nicholas Frontiere, Cody D. Raskin, J. Michael Owen
        "CRKSPH - A Conservative Reproducing Kernel Smoothed Particle
        Hydrodynamics Scheme", Journal of Computational Physics 332 (2017),
        pp. 160--209.

'''

from math import sqrt, exp
from compyle.api import declare
from pysph.sph.equation import Equation, Group, MultiStageEquations
from pysph.sph.wc.linalg import (
    augmented_matrix, dot, gj_solve, identity, mat_vec_mult
)
from pysph.sph.scheme import Scheme
from pysph.base.utils import get_particle_array
from pysph.sph.integrator import Integrator, IntegratorStep
from pysph.base.particle_array import get_ghost_tag

# Constants
GHOST_TAG = get_ghost_tag()


class CRKSPHPreStep(Equation):

    def __init__(self, dest, sources, dim=2):
        self.dim = dim
        super(CRKSPHPreStep, self).__init__(dest, sources)

    def _get_helpers_(self):
        return [augmented_matrix, gj_solve, identity, dot, mat_vec_mult]

    def loop_all(self, d_idx, d_x, d_y, d_z, d_h, s_x, s_y, s_z, s_h, s_m,
                 s_rho, SPH_KERNEL, NBRS, N_NBRS, d_ai, d_gradai, d_bi, s_V,
                 d_gradbi):
        x = d_x[d_idx]
        y = d_y[d_idx]
        z = d_z[d_idx]
        h = d_h[d_idx]
        i, j, k, s_idx, d, d2 = declare('int', 6)
        alp, bet, gam, phi, psi = declare('int', 5)
        xij = declare('matrix(3)')
        dwij = declare('matrix(3)')
        d = self.dim
        d2 = d*d

        m0 = 0.0
        m1 = declare('matrix(3)')
        m2 = declare('matrix(9)')
        temp_vec = declare('matrix(3)')
        temp_aug_m2 = declare('matrix(18)')
        m2inv = declare('matrix(9)')
        grad_m0 = declare('matrix(3)')
        grad_m1 = declare('matrix(9)')
        grad_m2 = declare('matrix(27)')
        ai = 0.0
        bi = declare('matrix(3)')
        grad_ai = declare('matrix(3)')
        grad_bi = declare('matrix(9)')

        for i in range(3):
            m1[i] = 0.0
            grad_m0[i] = 0.0
            bi[i] = 0.0
            grad_ai[i] = 0.0
            for j in range(3):
                m2[3*i + j] = 0.0
                grad_m1[3*i + j] = 0.0
                grad_bi[3*i + j] = 0.0
                for k in range(3):
                    grad_m2[9*i + 3*j + k] = 0.0
        for i in range(N_NBRS):
            s_idx = NBRS[i]
            xij[0] = x - s_x[s_idx]
            xij[1] = y - s_y[s_idx]
            xij[2] = z - s_z[s_idx]
            hij = (h + s_h[s_idx]) * 0.5
            rij = sqrt(xij[0] * xij[0] + xij[1] * xij[1] + xij[2] * xij[2])
            wij = SPH_KERNEL.kernel(xij, rij, hij)
            SPH_KERNEL.gradient(xij, rij, hij, dwij)
            V = 1.0/s_V[s_idx]

            m0 += V * wij
            for alp in range(d):
                m1[alp] += V * wij * xij[alp]
                for bet in range(d):
                    m2[d*alp + bet] += V * wij * xij[alp] * xij[bet]
            for gam in range(d):
                grad_m0[gam] += V * dwij[gam]
                for alp in range(d):
                    fac = 1.0 if alp == gam else 0.0
                    temp = (xij[alp] * dwij[gam] + fac * wij)
                    grad_m1[d*gam + alp] += V * temp
                    for bet in range(d):
                        fac2 = 1.0 if bet == gam else 0.0
                        temp = xij[alp] * fac2 + xij[bet] * fac
                        temp2 = (xij[alp] * xij[bet] * dwij[gam] + temp * wij)
                        grad_m2[d2*gam + d*alp + bet] += V * temp2

        identity(m2inv, d)
        augmented_matrix(m2, m2inv, d, d, d, temp_aug_m2)

        # If is_singular > 0 then matrix was singular
        is_singular = gj_solve(temp_aug_m2, d, d, m2inv)

        if is_singular > 0.0:
            # Cannot do much if the matrix is singular.  Perhaps later
            # we can tag such particles to see if the user can do something.
            pass
        else:
            mat_vec_mult(m2inv, m1, d, temp_vec)

            # Eq. 12.
            ai = 1.0/(m0 - dot(temp_vec, m1, d))
            # Eq. 13.
            mat_vec_mult(m2inv, m1, d, bi)
            for gam in range(d):
                bi[gam] = -bi[gam]

            # Eq. 14. and 15.
            for gam in range(d):
                temp1 = grad_m0[gam]
                for alp in range(d):
                    temp2 = 0.0
                    for bet in range(d):
                        temp1 -= m2inv[d*alp + bet] * (
                            m1[bet] * grad_m1[d*gam + alp] +
                            m1[alp] * grad_m1[d*gam + bet]
                        )
                        temp2 -= (
                            m2inv[d*alp + bet] * grad_m1[d*gam + bet]
                        )
                        for phi in range(d):
                            for psi in range(d):
                                temp1 += (
                                    m2inv[d*alp + phi] * m2inv[d*psi + bet] *
                                    grad_m2[d2*gam + d*phi + psi] *
                                    m1[bet] * m1[alp]
                                )
                                temp2 += (
                                    m2inv[d*alp + phi] * m2inv[d*psi + bet] *
                                    grad_m2[d2*gam + d*phi + psi] * m1[bet]
                                )
                    grad_bi[d*gam + alp] = temp2
                grad_ai[gam] = -ai*ai*temp1

        if N_NBRS < 2 or is_singular > 0.0:
            d_ai[d_idx] = 1.0
            for i in range(d):
                d_gradai[d * d_idx + i] = 0.0
                d_bi[d * d_idx + i] = 0.0
                for j in range(d):
                    d_gradbi[d2 * d_idx + d * i + j] = 0.0
        else:
            d_ai[d_idx] = ai
            for i in range(d):
                d_gradai[d * d_idx + i] = grad_ai[i]
                d_bi[d * d_idx + i] = bi[i]
                for j in range(d):
                    d_gradbi[d2 * d_idx + d * i + j] = grad_bi[d*i + j]


class CRKSPH(Equation):
    r"""**Conservative Reproducing Kernel SPH**

    Equations from the paper [CRKSPH2017].

    .. math::
            W_{ij}^{R} = A_{i}\left(1+B_{i}^{\alpha}x_{ij}^{\alpha}
            \right)W_{ij}
    .. math::
            \partial_{\gamma}W_{ij}^{R} = A_{i}\left(1+B_{i}^{\alpha}
            x_{ij}^{\alpha}\right)\partial_{\gamma}W_{ij} +
            \partial_{\gamma}A_{i}\left(1+B_{i}^{\alpha}x_{ij}^{\alpha}
            \right)W_{ij} + A_{i}\left(\partial_{\gamma}B_{i}^{\alpha}
            x_{ij}^{\alpha} + B_{i}^{\gamma}\right)W_{ij}
    .. math::
            \nabla\tilde{W}_{ij} = 0.5 * \left(\nabla W_{ij}^{R}-\nabla
             W_{ji}^{R} \right)

    where,

    .. math::
            A_{i} = \left[m_{0} - \left(m_{2}^{-1}\right)^{\alpha \beta}
            m_1^{\beta}m_1^{\alpha}\right]^{-1}
    .. math::
            B_{i}^{\alpha} = -\left(m_{2}^{-1}\right)^{\alpha \beta}
            m_{1}^{\beta}
    .. math::
            \partial_{\gamma}A_{i} = -A_{i}^{2}\left(\partial_{\gamma}
            m_{0}-\left(m_{2}^{-1}\right)^{\alpha \beta}\left(
            m_{1}^{\beta}\partial_{\gamma}m_{1}^{\alpha} +
            \partial_{\gamma}m_{1}^{\beta}m_{1}^{\alpha}\right) +
            \left(m_{2}^{-1}\right)^{\alpha \phi}\partial_{\gamma}
            m_{2}^{\phi \psi}\left(m_{2}^{-1}\right)^{\psi \beta}
            m_{1}^{\beta}m_{1}^{\alpha} \right)
    .. math::
            \partial_{\gamma}B_{i}^{\alpha} = -\left(m_{2}^{-1}\right)^
            {\alpha \beta}\partial_{\gamma}m_{1}^{\beta} +
            \left(m_{2}^{-1}\right)^
            {\alpha \phi}\partial_{\gamma}m_{2}^{\phi \psi}\left(m_{2}^
            {-1}\right)^{\psi \beta}m_{1}^{\beta}
    .. math::
            m_{0} = \sum_{j}V_{j}W_{ij}
    .. math::
            m_{1}^{\alpha} = \sum_{j}V_{j}x_{ij}^{\alpha}W_{ij}
    .. math::
            m_{2}^{\alpha \beta} = \sum_{j}V_{j}x_{ij}^{\alpha}
            x_{ij}^{\beta}W_{ij}
    .. math::
            \partial_{\gamma}m_{0} = \sum_{j}V_{j}\partial_{\gamma}
            W_{ij}
    .. math::
            \partial_{\gamma}m_{1}^{\alpha} = \sum_{j}V_{j}\left[
            x_{ij}^{\alpha}\partial_{\gamma}W_{ij}+\delta^
            {\alpha \gamma}W_{ij} \right]
    .. math::
            \partial_{\gamma}m_{2}^{\alpha \beta} = \sum_{j}V_{j}\left[
            x_{ij}^{\alpha}x_{ij}^{\beta}\partial_{\gamma}W_{ij} +
            \left(x_{ij}^{\alpha}\delta^{\beta \gamma} + x_{ij}^{\beta}
            \delta^{\alpha \gamma} \right)W_{ij} \right]
    """

    def __init__(self, dest, sources, dim=2, tol=0.5):
        r"""
        Parameters
        ----------

        dim : int
            Dimensionality of the problem.
        tol : float
            Tolerence value to decide std or corrected kernel
        """
        self.dim = dim
        self.tol = tol
        super(CRKSPH, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, d_ai, d_gradai, d_cwij, d_bi, d_gradbi,
             WIJ, DWIJ, XIJ, HIJ):
        alp, gam, d = declare('int', 3)
        res = declare('matrix(3)')
        dbxij = declare('matrix(3)')
        d = self.dim
        ai = d_ai[d_idx]
        eps = 1.0e-04 * HIJ
        bxij = 0.0
        for alp in range(d):
            bxij += d_bi[d*d_idx + alp] * XIJ[alp]
        for gam in range(d):
            temp = 0.0
            for alp in range(d):
                temp += d_gradbi[d*d*d_idx + d*gam + alp]*XIJ[alp]
            dbxij[gam] = temp

        d_cwij[d_idx] = (ai*(1 + bxij))

        for gam in range(d):
            res[gam] = ((ai * DWIJ[gam] + d_gradai[d * d_idx + gam] * WIJ) *
                        (1 + bxij))
            res[gam] += ai * (dbxij[gam] + d_bi[d * d_idx + gam]) * WIJ

        res_mag = 0.0
        dwij_mag = 0.0
        for i in range(d):
            res_mag += abs(res[i])
            dwij_mag += abs(DWIJ[i])
        change = abs(res_mag - dwij_mag)/(dwij_mag + eps)
        if change < self.tol:
            for i in range(d):
                DWIJ[i] = res[i]


class CRKSPHSymmetric(Equation):
    r"""**Conservative Reproducing Kernel SPH**

    This is symmetric and will only work for particles of the same array.

    Equations from the paper [CRKSPH2017].

    .. math::
            W_{ij}^{R} = A_{i}\left(1+B_{i}^{\alpha}x_{ij}^{\alpha}
            \right)W_{ij}
    .. math::
            \partial_{\gamma}W_{ij}^{R} = A_{i}\left(1+B_{i}^{\alpha}
            x_{ij}^{\alpha}\right)\partial_{\gamma}W_{ij} +
            \partial_{\gamma}A_{i}\left(1+B_{i}^{\alpha}x_{ij}^{\alpha}
            \right)W_{ij} + A_{i}\left(\partial_{\gamma}B_{i}^{\alpha}
            x_{ij}^{\alpha} + B_{i}^{\gamma}\right)W_{ij}
    .. math::
            \nabla\tilde{W}_{ij} = 0.5 * \left(\nabla W_{ij}^{R}-\nabla
             W_{ji}^{R} \right)

    where,

    .. math::
            A_{i} = \left[m_{0} - \left(m_{2}^{-1}\right)^{\alpha \beta}
            m1_{\beta}m1_{\alpha}\right]^{-1}
    .. math::
            B_{i}^{\alpha} = -\left(m_{2}^{-1}\right)^{\alpha \beta}
            m_{1}^{\beta}
    .. math::
            \partial_{\gamma}A_{i} = -A_{i}^{2}\left(\partial_{\gamma}
            m_{0}-\left[m_{2}^{-1}\right]^{\alpha \beta}\left[
            m_{1}^{\beta}\partial_{\gamma}m_{1}^{\beta}m_{1}^{\alpha} +
            \partial_{\gamma}m_{1}^{\alpha}m_{1}^{\beta}\right] +
            \left[m_{2}^{-1}\right]^{\alpha \phi}\partial_{\gamma}
            m_{2}^{\phi \psi}\left[m_{2}^{-1}\right]^{\psi \beta}
            m_{1}^{\beta}m_{1}^{\alpha} \right)
    .. math::
            \partial_{\gamma}B_{i}^{\alpha} = -\left(m_{2}^{-1}\right)^ {\alpha
            \beta}\partial_{\gamma}m_{1}^{\beta} + \left(m_{2}^{-1}\right)^
            {\alpha \phi}\partial_{\gamma}m_{2}^{\phi \psi}\left(m_{2}^
            {-1}\right)^{\psi \beta}m_{1}^{\beta}
    .. math::
            m_{0} = \sum_{j}V_{j}W_{ij}
    .. math::
            m_{1}^{\alpha} = \sum_{j}V_{j}x_{ij}^{\alpha}W_{ij}
    .. math::
            m_{2}^{\alpha \beta} = \sum_{j}V_{j}x_{ij}^{\alpha}
            x_{ij}^{\beta}W_{ij}
    .. math::
            \partial_{\gamma}m_{0} = \sum_{j}V_{j}\partial_{\gamma}
            W_{ij}
    .. math::
            \partial_{\gamma}m_{1}^{\alpha} = \sum_{j}V_{j}\left[
            x_{ij}^{\alpha}\partial_{\gamma}W_{ij}+\delta^
            {\alpha \gamma}W_{ij} \right]
    .. math::
            \partial_{\gamma}m_{2}^{\alpha \beta} = \sum_{j}V_{j}\left[
            x_{ij}^{\alpha}x_{ij}^{\beta}\partial_{\gamma}W_{ij} +
            \left(x_{ij}^{\alpha}\delta^{\beta \gamma} + x_{ij}^{\beta}
            \delta^{\alpha \gamma} \right)W_{ij} \right]
    """

    def __init__(self, dest, sources, dim=2, tol=0.5):
        self.dim = dim
        self.tol = tol
        super(CRKSPHSymmetric, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, d_ai, d_gradai, d_cwij, d_bi, d_gradbi, s_ai,
             s_gradai, s_bi, s_gradbi, d_h, s_h, WIJ, DWIJ, XIJ, HIJ, RIJ,
             DWI, DWJ, WI, WJ, SPH_KERNEL):
        alp, gam, d = declare('int', 3)
        dbxij = declare('matrix(3)')
        dbxji = declare('matrix(3)')
        dwij, dwji = declare('matrix(3)', 2)
        SPH_KERNEL.gradient(XIJ, RIJ, d_h[d_idx], dwij)
        SPH_KERNEL.gradient(XIJ, RIJ, s_h[s_idx], dwji)
        wij = SPH_KERNEL.kernel(XIJ, RIJ, d_h[d_idx])
        wji = SPH_KERNEL.kernel(XIJ, RIJ, s_h[s_idx])
        d = self.dim
        ai = d_ai[d_idx]
        aj = s_ai[s_idx]
        bxij = 0.0
        bxji = 0.0
        for alp in range(d):
            bxij += d_bi[d*d_idx + alp] * XIJ[alp]
            bxji -= s_bi[d*s_idx + alp] * XIJ[alp]
        for gam in range(d):
            temp = 0.0
            temp1 = 0.0
            for alp in range(d):
                temp += d_gradbi[d*d*d_idx + d*gam + alp]*XIJ[alp]
                temp1 -= s_gradbi[d*d*s_idx + d*gam + alp]*XIJ[alp]
            dbxij[gam] = temp
            dbxji[gam] = temp1

        d_cwij[d_idx] = (ai*(1 + bxij))
        WI = (ai*(1 + bxij)) * WIJ
        WJ = (aj*(1 + bxji)) * WIJ

        for gam in range(d):
            temp = ((ai * dwij[gam] + d_gradai[d * d_idx + gam] * wij) *
                    (1 + bxij))
            temp += ai * (dbxij[gam] + d_bi[d * d_idx + gam]) * wij
            temp1 = ((-aj * dwji[gam] + s_gradai[d * s_idx + gam] * wji) *
                     (1 + bxji))
            temp1 += aj * (dbxji[gam] + s_bi[d * s_idx + gam]) * wji
            DWIJ[gam] = 0.5*(temp - temp1)
            DWI[gam] = temp
            DWJ[gam] = temp1


class NumberDensity(Equation):
    r"""**Number Density**

    From [CRKSPH2017], equation (75):

    .. math::
        V_{i}^{-1} = \sum_{j} W_{i}

    Note that the quantity `V` is the inverse of particle volume, so when using
    in the equation use :math:`\frac{1}{V}`.
    """
    def initialize(self, d_idx, d_V):
        d_V[d_idx] = 0.0

    def loop(self, d_idx, d_V, WI):
        d_V[d_idx] += WI


class SummationDensityCRKSPH(Equation):
    r"""**Summation Density CRKSPH**

    From [CRKSPH2017], equation (76):

    .. math::
        \rho_{i} = \frac{\sum_j m_{ij} V_j W_{ij}^R}
        {\sum_{j} V_{j}^{2} W_{ij}^R}

    where,

    .. math::
        mij = \begin{cases}
        m_j, &i \text{ and } j \text{ are same material} \\
        m_i, &i \text{ and } j \text{ are different material}
        \end{cases}

    Note that in this equation we are just using :math:`m_{ij}` to be
    :math:`m_i` as the mass remains constant through out the simulation.
    """

    def initialize(self, d_idx, d_rho, d_rhofac):
        d_rho[d_idx] = 0.0
        d_rhofac[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, d_rho, d_rhofac, s_V, WIJ, d_cwij):
        Vj = 1.0/s_V[s_idx]
        fac = Vj * d_cwij[d_idx] * WIJ
        d_rho[d_idx] += d_m[d_idx] * fac
        d_rhofac[d_idx] += Vj * fac

    def post_loop(self, d_idx, d_rho, d_rhofac):
        d_rho[d_idx] = d_rho[d_idx] / d_rhofac[d_idx]


class VelocityGradient(Equation):
    r"""**Velocity Gradient**

    From [CRKSPH2017], equation (74)

    .. math::
        \partial_{\beta} v_i^{\alpha} = -\sum_j V_j v_{ij}^{\alpha}
        \partial_{\beta} W_{ij}^R
    """
    def __init__(self, dest, sources, dim):
        r"""
        Parameters
        ----------

        dim : int
            Dimensionality of the Problem.
        """
        self.dim = dim
        super(VelocityGradient, self).__init__(dest, sources)

    def initialize(self, d_idx, d_gradv):
        i = declare('int')
        for i in range(9):
            d_gradv[9*d_idx + i] = 0.0

    def loop(self, d_idx, s_idx, s_V, d_gradv, d_h, s_h, XIJ, DWIJ, VIJ, DWI):
        alp, bet, d = declare('int', 3)
        d = self.dim

        Vj = 1.0/s_V[s_idx]

        for alp in range(d):
            for bet in range(d):
                d_gradv[d_idx*d*d + d*alp + bet] += -Vj * VIJ[alp] * DWI[bet]


class MomentumEquation(Equation):
    r"""**Momentum Equation**

    From [CRKSPH2017], equation (64):

    .. math::
        \frac{Dv_{i}^{\alpha}}{Dt} = -\frac{1}{2 m_i}\sum_{j} V_i V_j (P_i
        + P_j + Q_i + Q_j)
        (\partial_{\alpha}W_{ij}^R - \partial_{\alpha} W_{ji}^R)

    where,

    .. math::
        V_{i/j} = \text{dest/source particle number density}

    .. math::
        P_{i/j} = \text{dest/source particle pressure}

    .. math::
        Q_i = \rho_{i} (-C_{l} c_{i} \mu_{i} + C_{q} \mu_{i}^{2})

    .. math::
        \mu_i = \min \left(0, \frac{\hat{v}_{ij}
        \eta_{i}^{\alpha}}{\eta_{i}^{\alpha}\eta_{i}^{\alpha}
        + \epsilon^{2}}\right)

    .. math::
        \hat{v}_{ij}^{\alpha} = v_{i}^{\alpha} - v_{j}^{\alpha}
        - \frac{\phi_{ij}}{2}\left(\partial_{\beta} v_i^{\alpha}
        + \partial_{\beta}v_j^{\alpha}\right) x_{ij}^{\beta}

    .. math::
        \phi_{ij} = \max \left[0, \min \left[1, \frac{4r_{ij}}{(1
        + r_{ij})^2}\right]\right] \times
        \begin{cases}
        \exp{\left[-\left((\eta_{ij}
        - \eta_{crit})/\eta_{fold}\right)^2\right]}, &\eta_{ij} < \eta_{crit}
        \\
        1, &  \eta_{ij} >= \eta_{crit}
        \end{cases}

    .. math::
        \eta_{ij} = \min(\eta_i, \eta_j)

    .. math::
        \eta_{i/j} = (x_{ij}^{\alpha} x_{ij}^{\alpha})^{1/2} / h_{i/j}

    .. math::
        r_{ij} = \frac{\partial_{\beta} v_i^{\alpha} x_{ij}^{\alpha}
        x_{ij}^{\beta}}{\partial_{\beta} v_j^{\alpha}x_{ij}^{\alpha}
        x_{ij}^{\beta}}

    .. math::
        \partial_{\beta} v_i^{\alpha} = -\sum_j V_j v_{ij}^{\alpha}
        \partial_{\beta} W_{ij}^R
    """

    def __init__(self, dest, sources, dim, gx=0.0, gy=0.0, gz=0.0, cl=2, cq=1,
                 eta_crit=0.3, eta_fold=0.2, tol=0.5):
        self.dim = dim
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.cl = cl
        self.cq = cq
        self.eta_crit = eta_crit
        self.eta_fold = eta_fold
        self.tol = tol
        super(MomentumEquation, self).__init__(dest, sources)

    def _get_helpers_(self):
        return [dot]

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = self.gx
        d_av[d_idx] = self.gy
        d_aw[d_idx] = self.gz

    def loop(self, d_idx, s_idx, d_m, s_m, d_rho, s_rho, d_p, s_p, d_cs, s_cs,
             d_u, d_v, d_w, s_u, s_v, s_w, d_gradv, s_gradv, d_h, s_h, s_ai,
             s_bi, s_gradai, s_gradbi, d_au, d_av, d_aw, d_V, s_V, XIJ, VIJ,
             DWIJ, EPS, HIJ, WIJ, DWI, DWJ):
        Cl = self.cl
        Cq = self.cq
        eta_fold = self.eta_fold
        eta_crit = self.eta_crit

        mi = d_m[d_idx]
        Vi = 1.0/d_V[d_idx]
        Vj = 1.0/s_V[s_idx]

        pi = d_p[d_idx]
        pj = s_p[s_idx]

        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]

        ci = d_cs[d_idx]
        cj = s_cs[s_idx]

        hi = d_h[d_idx]
        hj = s_h[s_idx]

        alp, bet, i, d = declare('int', 4)
        uijhat, tmpdvxij = declare('matrix(3)', 2)

        d = self.dim

        tmpri = 0.0
        tmprj = 0.0
        for alp in range(d):
            for bet in range(d):
                tmpri += d_gradv[d*d*d_idx + d*alp + bet] * XIJ[alp] * XIJ[bet]
                tmprj += s_gradv[d*d*s_idx + d*alp + bet] * XIJ[alp] * XIJ[bet]
        rij = tmpri/tmprj

        tmprij = min(1, 4*rij/((1 + rij)*(1 + rij)))
        phiij = max(0, tmprij)

        tmpxij = dot(XIJ, XIJ, d)
        tmpxij2 = sqrt(tmpxij)
        etai_scalar = tmpxij2/hi
        etaj_scalar = tmpxij2/hj
        etaij = min(etai_scalar, etaj_scalar)

        if etaij < eta_crit:
            tmpphi = (etaij - eta_crit)/eta_fold
            phiij = phiij * exp(-tmpphi*tmpphi)

        for alp in range(d):
            s = 0.0
            for bet in range(d):
                s += (d_gradv[d*d*d_idx + d*alp + bet] +
                      s_gradv[d*d*s_idx + d*alp + bet]) * XIJ[bet]
            tmpdvxij[alp] = s

        uijhat[0] = d_u[d_idx] - s_u[s_idx] - 0.5*phiij * tmpdvxij[0]
        uijhat[1] = d_v[d_idx] - s_v[s_idx] - 0.5*phiij * tmpdvxij[1]
        uijhat[2] = d_w[d_idx] - s_w[s_idx] - 0.5*phiij * tmpdvxij[2]

        tmpmui = dot(uijhat, XIJ, d) / (tmpxij/hi + EPS * hi)
        mui = min(0, tmpmui)

        tmpmuj = dot(uijhat, XIJ, d) / (tmpxij/hi + EPS * hj)
        muj = min(0, tmpmuj)

        Qi = rhoi * (-Cl*ci*mui + Cq*mui*mui)
        Qj = rhoj * (-Cl*cj*muj + Cq*muj*muj)

        fac = -(1.0/mi) * Vi * Vj * (pi + pj + Qi + Qj)
        d_au[d_idx] += fac * DWIJ[0]
        d_av[d_idx] += fac * DWIJ[1]
        d_aw[d_idx] += fac * DWIJ[2]


class EnergyEquation(Equation):
    r"""**Energy Equation**

    From [CRKSPH2017], equation (66):

    .. math::
        \Delta u_{ij} = \frac{f_{ij}}{2}\left(v_j^{\alpha}(t) + v_j^{\alpha}(t
        + \Delta t) - v_i^{\alpha}(t) - v_i^{\alpha}(t + \Delta t)\right)
        \frac{Dv_{ij}^{\alpha}}{Dt}

    .. math::
        f_{ij} = \begin{cases}
        1/2 &|s_i - s_j| = 0,\\
        s_{\min} / (s_{\min} + s_{\max}) &\Delta u_{ij}\times(s_i - s_j) > 0\\
        s_{\max} / (s_{\min} + s_{\max}) &\Delta u_{ij}\times(s_i - s_j) < 0\\
        \end{cases}

    .. math::
        s_{\min} = \min(|s_i|, |s_j|)

    .. math::
        s_{\max} = \max(|s_i|, |s_j|)

    .. math::
        s_{i/j} = \frac{p_{i/j}}{\rho_{i/j}^\gamma}

    see MomentumEquation for :math:`\frac{Dv_{ij}^{\alpha}}{Dt}`
    """

    def __init__(self, dest, sources, dim, gamma, gx=0.0, gy=0.0, gz=0.0,
                 cl=2, cq=1, eta_crit=0.5, eta_fold=0.2, tol=0.5):
        self.dim = dim
        self.gamma = gamma
        self.dim = dim
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.cl = cl
        self.cq = cq
        self.eta_crit = eta_crit
        self.eta_fold = eta_fold
        self.tol = tol
        super(EnergyEquation, self).__init__(dest, sources)

    def _get_helpers_(self):
        return [dot]

    def initialize(self, d_idx, d_ae):
        d_ae[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_au, d_av, d_aw, d_ae, s_au, s_av, s_aw,
             d_u0, d_v0, d_w0, s_u0, s_v0, s_w0, d_u, d_v, d_w, s_u, s_v, s_w,
             d_p, d_rho, s_p, s_rho, d_m, d_V, s_V, d_cs, s_cs, d_h, s_h,
             XIJ, d_gradv, s_gradv, EPS, DWIJ):
        Cl = self.cl
        Cq = self.cq
        eta_fold = self.eta_fold
        eta_crit = self.eta_crit

        mi = d_m[d_idx]
        Vi = 1.0/d_V[d_idx]
        Vj = 1.0/s_V[s_idx]

        pi = d_p[d_idx]
        pj = s_p[s_idx]

        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]

        ci = d_cs[d_idx]
        cj = s_cs[s_idx]

        hi = d_h[d_idx]
        hj = s_h[s_idx]

        alp, bet, i, d = declare('int', 4)
        uijhat, tmpdvxij = declare('matrix(3)', 2)

        d = self.dim

        tmpri = 0.0
        tmprj = 0.0
        for alp in range(d):
            for bet in range(d):
                tmpri += d_gradv[d*d*d_idx + d*alp + bet] * XIJ[alp] * XIJ[bet]
                tmprj += s_gradv[d*d*s_idx + d*alp + bet] * XIJ[alp] * XIJ[bet]
        rij = tmpri/tmprj

        tmprij = min(1, 4*rij/((1 + rij)*(1 + rij)))
        phiij = max(0, tmprij)

        tmpxij = dot(XIJ, XIJ, d)
        tmpxij2 = sqrt(tmpxij)
        etai_scalar = tmpxij2/hi
        etaj_scalar = tmpxij2/hj
        etaij = min(etai_scalar, etaj_scalar)

        if etaij < eta_crit:
            tmpphi = (etaij - eta_crit)/eta_fold
            phiij = phiij * exp(-tmpphi*tmpphi)

        for alp in range(d):
            s = 0.0
            for bet in range(d):
                s += (d_gradv[d*d*d_idx + d*alp + bet] +
                      s_gradv[d*d*s_idx + d*alp + bet]) * XIJ[bet]
            tmpdvxij[alp] = s

        uijhat[0] = d_u0[d_idx] - s_u0[s_idx] - 0.5*phiij * tmpdvxij[0]
        uijhat[1] = d_v0[d_idx] - s_v0[s_idx] - 0.5*phiij * tmpdvxij[1]
        uijhat[2] = d_w0[d_idx] - s_w0[s_idx] - 0.5*phiij * tmpdvxij[2]

        tmpmui = dot(uijhat, XIJ, d) / (tmpxij/hi + EPS * hi)
        mui = min(0, tmpmui)

        tmpmuj = dot(uijhat, XIJ, d) / (tmpxij/hi + EPS * hj)
        muj = min(0, tmpmuj)

        Qi = rhoi * (-Cl*ci*mui + Cq*mui*mui)
        Qj = rhoj * (-Cl*cj*muj + Cq*muj*muj)

        fac = -(1.0 / mi) * Vi * Vj * (pi + pj + Qi + Qj)

        gamma = self.gamma
        d = self.dim
        auij, delu = declare('matrix(3)', 2)
        auij[0] = fac * DWIJ[0]
        auij[1] = fac * DWIJ[1]
        auij[2] = fac * DWIJ[2]

        delu[0] = s_u0[s_idx] + s_u[s_idx] - d_u0[d_idx] - d_u[d_idx]
        delu[1] = s_v0[s_idx] + s_v[s_idx] - d_v0[d_idx] - d_v[d_idx]
        delu[2] = s_w0[s_idx] + s_w[s_idx] - d_w0[d_idx] - d_w[d_idx]

        aeij = dot(delu, auij, d)

        si = d_p[d_idx]/((d_rho[d_idx])**gamma)
        sj = s_p[s_idx]/((s_rho[s_idx])**gamma)
        smin = min(abs(si), abs(sj))
        smax = max(abs(si), abs(sj))

        fij = 0.5
        sdiff = si - sj
        if sdiff * aeij > 0:
            fij = smin/(smin + smax)
        elif sdiff * aeij < 0:
            fij = smax/(smin + smax)

        d_ae[d_idx] += 0.5*fij * aeij


class StateEquation(Equation):
    r"""**State Equation**

    State equation for ideal gas, from [CRKSPH2017] equation (77):

    .. math::
        p_i = (\gamma - 1)\rho_{i} u_i

    where, :math:`u_i` is the specific thermal energy.
    """
    def __init__(self, dest, sources, gamma):
        self.gamma = gamma
        super(StateEquation, self).__init__(dest, sources)

    def initialize(self, d_idx, d_p, d_rho, d_e):
        d_p[d_idx] = (self.gamma - 1) * d_rho[d_idx] * d_e[d_idx]


class SpeedOfSound(Equation):
    def __init__(self, dest, sources=None, gamma=7.0):
        super(SpeedOfSound, self).__init__(dest, sources)
        self.gamma = gamma

    def initialize(self, d_cs, d_idx, d_p, d_rho):
        d_cs[d_idx] = (self.gamma * d_p[d_idx] / d_rho[d_idx])**0.5


class CRKSPHUpdateGhostProps(Equation):
    def __init__(self, dest, sources=None, dim=2):
        super(CRKSPHUpdateGhostProps, self).__init__(dest, sources)
        self.dim = dim
        assert GHOST_TAG == 2

    def initialize(self, d_idx, d_orig_idx, d_p, d_cs, d_V, d_ai, d_bi, d_tag,
                   d_gradbi, d_gradai, d_rho, d_u0, d_v0, d_w0, d_gradv, d_u,
                   d_v, d_w):
        idx, d, alp, bet = declare('int', 4)
        d = self.dim
        if d_tag[d_idx] == 2:
            idx = d_orig_idx[d_idx]
            d_p[d_idx] = d_p[idx]
            d_cs[d_idx] = d_cs[idx]
            d_V[d_idx] = d_V[idx]
            d_rho[d_idx] = d_rho[idx]
            d_u0[d_idx] = d_u0[idx]
            d_v0[d_idx] = d_v0[idx]
            d_w0[d_idx] = d_w0[idx]
            d_u[d_idx] = d_u[idx]
            d_v[d_idx] = d_v[idx]
            d_w[d_idx] = d_w[idx]
            d_ai[d_idx] = d_ai[idx]
            for alp in range(d):
                d_gradai[d*d_idx + alp] = d_gradai[d*idx + alp]
                d_bi[d*d_idx + alp] = d_bi[d*idx + alp]
                for bet in range(d):
                    d_gradv[d*d*d_idx + d*alp + bet] = \
                        d_gradv[d*d*idx + d*alp + bet]
                    d_gradbi[d*d*d_idx + d*alp + bet] = \
                        d_gradbi[d*d*idx + d*alp + bet]


def get_particle_array_crksph(constants=None, **props):
    crksph_props = [
        'e', 'au', 'av', 'aw', 'ae', 'u0', 'v0', 'w0', 'cs', 'V', 'rhofac',
        'x0', 'y0', 'z0', 'rho0', 'ax', 'ay', 'az', 'arho'
    ]

    pa = get_particle_array(
        additional_props=crksph_props, constants=constants, **props
    )
    pa.add_property('cwij')
    pa.add_property('ai')
    pa.add_property('bi', stride=3)
    pa.add_property('gradai', stride=3)
    pa.add_property('gradbi', stride=9)
    pa.add_property('gradv', stride=9)
    pa.add_output_arrays(['p', 'V'])
    return pa


class CRKSPHIntegrator(Integrator):
    def one_timestep(self, t, dt):
        self.stage1()
        self.do_post_stage(dt, 1)

        self.compute_accelerations(0)

        self.stage2()
        self.do_post_stage(dt, 2)

        self.compute_accelerations(1)

        # We update domain here alone as positions only change here.
        self.stage3()
        self.do_post_stage(dt, 3)
        self.update_domain()


class CRKSPHStep(IntegratorStep):
    def stage1(self, d_idx, d_u, d_v, d_w, d_u0, d_v0, d_w0):
        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

    def stage2(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, dt):
        d_u[d_idx] += d_au[d_idx] * dt
        d_v[d_idx] += d_av[d_idx] * dt
        d_w[d_idx] += d_aw[d_idx] * dt

    def stage3(self, d_idx, d_e, d_ae, d_u, d_v, d_w, d_u0, d_v0, d_w0,
               d_x, d_y, d_z, dt):
        d_e[d_idx] += d_ae[d_idx] * dt
        d_x[d_idx] += 0.5*dt*(d_u[d_idx] + d_u0[d_idx])
        d_y[d_idx] += 0.5*dt*(d_v[d_idx] + d_v0[d_idx])
        d_z[d_idx] += 0.5*dt*(d_w[d_idx] + d_w0[d_idx])


class CRKSPHScheme(Scheme):
    def __init__(self, fluids, dim, rho0, c0, nu, h0, p0, gx=0.0,
                 gy=0.0, gz=0.0, cl=2, cq=1, gamma=7.0, eta_crit=0.3,
                 eta_fold=0.2, tol=0.5, has_ghosts=False):
        """
        Parameters
        ----------

        fluids: list
            a list with names of fluid particle arrays
        solids: list
            a list with names of solid (or boundary) particle arrays

        """
        self.fluids = fluids
        self.solver = None
        self.dim = dim
        self.rho0 = rho0
        self.c0 = c0
        self.h0 = h0
        self.p0 = p0
        self.nu = nu
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.cl = cl
        self.cq = cq
        self.gamma = gamma
        self.eta_crit = eta_crit
        self.eta_fold = eta_fold
        self.tol = tol
        self.has_ghosts = has_ghosts

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        Parameters
        ----------

        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import QuinticSpline
        if kernel is None:
            kernel = QuinticSpline(dim=self.dim)
        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        step_cls = CRKSPHStep
        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = step_cls()

        if integrator_cls is not None:
            cls = integrator_cls
            print("warning: CRKSPHIntegrator is not being used")
        else:
            cls = CRKSPHIntegrator
        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel,
            **kw
        )

    def get_equations(self):
        from pysph.sph.wc.viscosity import LaminarViscosity
        all = self.fluids

        equations_stage1 = []
        equations_stage2 = []

        eq00 = []
        for fluid in self.fluids:
            eq00.append(
                StateEquation(dest=fluid, sources=None, gamma=self.gamma)
            )
            eq00.append(
                SpeedOfSound(dest=fluid, sources=None, gamma=self.gamma)
            )
        equations_stage1.append(Group(equations=eq00))

        gh = []
        for fluid in self.fluids:
            gh.append(
                CRKSPHUpdateGhostProps(dest=fluid, sources=None, dim=self.dim)
            )
        equations_stage1.append(Group(equations=gh, real=False))

        eq0 = []
        for fluid in self.fluids:
            eq0.append(NumberDensity(dest=fluid, sources=all))
        equations_stage1.append(Group(equations=eq0, real=False))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(
                    CRKSPHUpdateGhostProps(
                        dest=fluid, sources=None, dim=self.dim
                        )
                )
            equations_stage1.append(Group(equations=gh, real=False))

        eq1 = []
        for fluid in self.fluids:
            eq1.append(CRKSPHPreStep(dest=fluid, sources=all, dim=self.dim))
        equations_stage1.append(Group(equations=eq1, real=False))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(
                    CRKSPHUpdateGhostProps(
                        dest=fluid, sources=None, dim=self.dim
                        )
                )
            equations_stage1.append(Group(equations=gh, real=False))

        eq2 = []
        for fluid in self.fluids:
            eq2.extend([
                CRKSPHSymmetric(
                    dest=fluid, sources=all, dim=self.dim, tol=self.tol
                    ),
                SummationDensityCRKSPH(dest=fluid, sources=all)
            ])
        equations_stage1.append(Group(equations=eq2, real=False))

        eq3 = []
        for fluid in self.fluids:
            eq3.append(
                StateEquation(dest=fluid, sources=None, gamma=self.gamma)
            )
            eq3.append(
                SpeedOfSound(dest=fluid, sources=None, gamma=self.gamma)
            )
        equations_stage1.append(Group(equations=eq3))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(
                    CRKSPHUpdateGhostProps(
                        dest=fluid, sources=None, dim=self.dim
                        )
                )
            equations_stage1.append(Group(equations=gh, real=False))

        eq4 = []
        for fluid in self.fluids:
            eq4.extend([
                CRKSPHSymmetric(
                    dest=fluid, sources=all, dim=self.dim, tol=self.tol
                    ),
                VelocityGradient(dest=fluid, sources=all, dim=self.dim)
            ])
        equations_stage1.append(Group(equations=eq4))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(
                    CRKSPHUpdateGhostProps(
                        dest=fluid, sources=None, dim=self.dim
                        )
                )
            equations_stage1.append(Group(equations=gh, real=False))

        eq5 = []
        for fluid in self.fluids:
            eq5.append(
                CRKSPHSymmetric(
                    dest=fluid, sources=all, dim=self.dim, tol=self.tol
                )
            )
            eq5.append(
                MomentumEquation(
                    dest=fluid, sources=all, dim=self.dim, gx=self.gx,
                    gy=self.gy, gz=self.gz, cl=self.cl, cq=self.cq,
                    eta_crit=self.eta_crit, eta_fold=self.eta_fold
                )
            )
            if abs(self.nu) > 1e-14:
                eq5.append(LaminarViscosity(
                    dest=fluid, sources=self.fluids, nu=self.nu
                ))
        equations_stage1.append(Group(equations=eq5))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(
                    CRKSPHUpdateGhostProps(
                        dest=fluid, sources=None, dim=self.dim
                        )
                )
            equations_stage2.append(Group(equations=gh, real=False))

        eq6 = []
        for fluid in self.fluids:
            eq6.append(
                CRKSPHSymmetric(
                    dest=fluid, sources=all, dim=self.dim, tol=self.tol
                    )
            )
            eq6.append(
                EnergyEquation(
                    dest=fluid, sources=all, dim=self.dim, gamma=self.gamma
                    )
            )
        equations_stage2.append(Group(equations=eq6))

        return MultiStageEquations([equations_stage1, equations_stage2])

    def setup_properties(self, particles, clean=True):
        import numpy
        particle_arrays = dict([(p.name, p) for p in particles])
        crksph_props = [
            'au', 'av', 'aw', 'ae', 'u0', 'v0', 'w0', 'cs', 'V', 'rhofac',
            'x0', 'y0', 'z0', 'rho0', 'ax', 'ay', 'az', 'arho'
        ]
        dummy = get_particle_array_crksph(name='junk')
        props = list(dummy.properties.keys())
        props += [dict(name=p, stride=v) for p, v in dummy.stride.items()]
        props += crksph_props
        output_props = dummy.output_property_arrays
        output_props += ['p', 'V', 'e']
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.add_property('cwij')
            pa.add_property('ai')
            pa.add_property('bi', stride=3)
            pa.add_property('gradai', stride=3)
            pa.add_property('gradbi', stride=9)
            pa.add_property('gradv', stride=9)
            pa.add_property('orig_idx', type='int')
            nfp = pa.get_number_of_particles()
            pa.orig_idx[:] = numpy.arange(nfp)
            pa.set_output_arrays(output_props)



"""
EDAC SPH formulation
#####################

Equations for the Entropically Damped Artificial Compressibility SPH scheme.

Please note that this scheme is still under development and this module may
change at some point in the future.

References
----------

    .. [PRKP2017] Prabhu Ramachandran and Kunal Puri, Entropically damped
       artificial compressibility for SPH, under review, 2017.
       http://arxiv.org/pdf/1311.2167v2.pdf

"""

from math import sin
from math import pi as M_PI

from pysph.base.utils import get_particle_array
from pysph.base.utils import DEFAULT_PROPS
from pysph.sph.equation import Equation, Group
from pysph.sph.integrator_step import IntegratorStep

from pysph.sph.scheme import Scheme, add_bool_argument


EDAC_PROPS = ('ap', 'au', 'av', 'aw', 'ax', 'ay', 'az',
              'x0', 'y0', 'z0', 'u0', 'v0', 'w0', 'p0', 'V')


def get_particle_array_edac(constants=None, **props):
    "Get the fluid array for the transport velocity formulation"

    pa = get_particle_array(
        constants=constants, additional_props=EDAC_PROPS, **props
    )
    pa.set_output_arrays(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'p',
                          'au', 'av', 'aw', 'ap', 'm', 'h'])

    return pa


EDAC_SOLID_PROPS = ('ap', 'p0', 'wij', 'uf', 'vf', 'wf', 'ug', 'vg', 'wg',
                    'ax', 'ay', 'az', 'V')


def get_particle_array_edac_solid(constants=None, **props):
    "Get the fluid array for the transport velocity formulation"

    pa = get_particle_array(
        constants=constants, additional_props=EDAC_SOLID_PROPS, **props
    )
    pa.set_output_arrays(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'p',
                          'h'])

    return pa


class ComputeAveragePressure(Equation):
    """Simple function to compute the average pressure at each particle.

    This is used for the Basa, Quinlan and Lastiwka correction from their 2009
    paper.  This equation should be in a separate group and computed before the
    Momentum equation.
    """
    def initialize(self, d_idx, d_pavg, d_nnbr):
        d_pavg[d_idx] = 0.0
        d_nnbr[d_idx] = 0.0

    def loop(self, d_idx, d_pavg, s_idx, s_p, d_nnbr):
        d_pavg[d_idx] += s_p[s_idx]
        d_nnbr[d_idx] += 1.0

    def post_loop(self, d_idx, d_pavg, d_nnbr):
        if d_nnbr[d_idx] > 0:
            d_pavg[d_idx] /= d_nnbr[d_idx]


class EDACStep(IntegratorStep):
    """Standard Predictor Corrector integrator for the WCSPH formulation

    Use this integrator for WCSPH formulations. In the predictor step,
    the particles are advanced to `t + dt/2`. The particles are then
    advanced with the new force computed at this position.

    This integrator can be used in PEC or EPEC mode.

    The same integrator can be used for other problems. Like for
    example solid mechanics (see SolidMechStep)

    """
    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_p0, d_p):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_p0[d_idx] = d_p[d_idx]

    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_p0, d_p, d_au, d_av,
               d_aw, d_ax, d_ay, d_az, d_ap, dt=0.0):
        dtb2 = 0.5*dt
        d_u[d_idx] = d_u0[d_idx] + dtb2*d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2*d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2*d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dtb2 * d_ax[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2 * d_ay[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2 * d_az[d_idx]

        d_p[d_idx] = d_p0[d_idx] + dtb2 * d_ap[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_p0, d_p, d_au, d_av,
               d_aw, d_ax, d_ay, d_az, d_ap, dt=0.0):

        d_u[d_idx] = d_u0[d_idx] + dt*d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt*d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt*d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dt * d_ax[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_ay[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_az[d_idx]

        d_p[d_idx] = d_p0[d_idx] + dt * d_ap[d_idx]


class SolidWallPressureBC(Equation):
    r"""Solid wall pressure boundary condition from Adami and Hu (transport
    velocity formulation).

    """
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz

        super(SolidWallPressureBC, self).__init__(dest, sources)

    def initialize(self, d_idx, d_p):
        d_p[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_p, s_p, s_rho,
             d_au, d_av, d_aw, WIJ, XIJ):

        # numerator of Eq. (27) ax, ay and az are the prescribed wall
        # accelerations which must be defined for the wall boundary
        # particle
        gdotxij = (self.gx - d_au[d_idx])*XIJ[0] + \
            (self.gy - d_av[d_idx])*XIJ[1] + \
            (self.gz - d_aw[d_idx])*XIJ[2]

        d_p[d_idx] += s_p[s_idx]*WIJ + s_rho[s_idx]*gdotxij*WIJ

    def post_loop(self, d_idx, d_wij, d_p):
        # extrapolated pressure at the ghost particle
        if d_wij[d_idx] > 1e-14:
            d_p[d_idx] /= d_wij[d_idx]


class ClampWallPressure(Equation):
    r"""Clamp the wall pressure to non-negative values.
    """
    def post_loop(self, d_idx, d_p):
        if d_p[d_idx] < 0.0:
            d_p[d_idx] = 0.0


class SourceNumberDensity(Equation):
    r"""Evaluates the number density due to the source particles"""
    def initialize(self, d_idx, d_wij):
        d_wij[d_idx] = 0.0

    def loop(self, d_idx, d_wij, WIJ):
        d_wij[d_idx] += WIJ


class SetWallVelocity(Equation):
    r"""Extrapolating the fluid velocity on to the wall Eq. (22) in REF1:

    .. math::

        \tilde{\boldsymbol{v}}_a = \frac{\sum_b\boldsymbol{v}_b W_{ab}}
        {\sum_b W_{ab}}

    Notes:

    This should be used only after (or in the same group) as the
    SolidWallPressureBC equation.

    The destination particle array for this equation should define the
    *filtered* velocity variables :math:`uf, vf, wf`.

    """
    def initialize(self, d_idx, d_uf, d_vf, d_wf):
        d_uf[d_idx] = 0.0
        d_vf[d_idx] = 0.0
        d_wf[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_uf, d_vf, d_wf,
             s_u, s_v, s_w, WIJ):
        # sum in Eq. (22)
        # this will be normalized in post loop
        d_uf[d_idx] += s_u[s_idx] * WIJ
        d_vf[d_idx] += s_v[s_idx] * WIJ
        d_wf[d_idx] += s_w[s_idx] * WIJ

    def post_loop(self, d_uf, d_vf, d_wf, d_wij, d_idx,
                  d_ug, d_vg, d_wg, d_u, d_v, d_w):

        # calculation is done only for the relevant boundary particles.
        # d_wij (and d_uf) is 0 for particles sufficiently away from the
        # solid-fluid interface

        # Note that d_wij is already computed for the pressure BC.
        if d_wij[d_idx] > 1e-12:
            d_uf[d_idx] /= d_wij[d_idx]
            d_vf[d_idx] /= d_wij[d_idx]
            d_wf[d_idx] /= d_wij[d_idx]

        # Dummy velocities at the ghost points using Eq. (23),
        # d_u, d_v, d_w are the prescribed wall velocities.
        d_ug[d_idx] = 2*d_u[d_idx] - d_uf[d_idx]
        d_vg[d_idx] = 2*d_v[d_idx] - d_vf[d_idx]
        d_wg[d_idx] = 2*d_w[d_idx] - d_wf[d_idx]


class NoSlipVelocityExtrapolation(Equation):
    '''No Slip boundary condition on the wall

    The velocity of the fluid is extrapolated over to the wall using
    shepard extrapolation. The velocity normal to the wall is reflected back
    to impose no penetration.
    '''
    def initialize(self, d_idx, d_u, d_v, d_w):
        d_u[d_idx] = 0.0
        d_v[d_idx] = 0.0
        d_w[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_u, d_v, d_w, s_u, s_v, s_w, WIJ,
             XIJ):
        d_u[d_idx] += s_u[s_idx]*WIJ
        d_v[d_idx] += s_v[s_idx]*WIJ
        d_w[d_idx] += s_w[s_idx]*WIJ

    def post_loop(self, d_idx, d_wij, d_u, d_v, d_w, d_xn, d_yn, d_zn):
        if d_wij[d_idx] > 1e-14:
            d_u[d_idx] /= d_wij[d_idx]
            d_v[d_idx] /= d_wij[d_idx]
            d_w[d_idx] /= d_wij[d_idx]

        projection = d_u[d_idx]*d_xn[d_idx] +\
            d_v[d_idx]*d_yn[d_idx] + d_w[d_idx]*d_zn[d_idx]

        d_u[d_idx] = d_u[d_idx] - 2 * projection * d_xn[d_idx]
        d_v[d_idx] = d_v[d_idx] - 2 * projection * d_yn[d_idx]
        d_w[d_idx] = d_w[d_idx] - 2 * projection * d_zn[d_idx]


class NoSlipAdvVelocityExtrapolation(Equation):
    '''No Slip boundary condition on the wall

    The advection velocity of the fluid is extrapolated over to the wall
    using shepard extrapolation. The advection velocity normal to the wall
    is reflected back to impose no penetration.
    '''
    def initialize(self, d_idx, d_uhat, d_vhat, d_what):
        d_uhat[d_idx] = 0.0
        d_vhat[d_idx] = 0.0
        d_what[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_uhat, d_vhat, d_what, s_uhat, s_vhat,
             s_what, WIJ, XIJ):
        d_uhat[d_idx] += s_uhat[s_idx]*WIJ
        d_vhat[d_idx] += s_vhat[s_idx]*WIJ
        d_what[d_idx] += s_what[s_idx]*WIJ

    def post_loop(self, d_idx, d_wij, d_uhat, d_vhat, d_what, d_xn, d_yn,
                  d_zn):
        if d_wij[d_idx] > 1e-14:
            d_uhat[d_idx] /= d_wij[d_idx]
            d_vhat[d_idx] /= d_wij[d_idx]
            d_what[d_idx] /= d_wij[d_idx]

        projection = d_uhat[d_idx]*d_xn[d_idx] +\
            d_vhat[d_idx]*d_yn[d_idx] + d_what[d_idx]*d_zn[d_idx]

        d_uhat[d_idx] = d_uhat[d_idx] - 2 * projection * d_xn[d_idx]
        d_vhat[d_idx] = d_vhat[d_idx] - 2 * projection * d_yn[d_idx]
        d_what[d_idx] = d_what[d_idx] - 2 * projection * d_zn[d_idx]


class MomentumEquation(Equation):
    r"""Momentum equation (gradient of pressure) based on the number
    density formulation of Hu and Adams JCP 213 (2006), 844-861.

    """
    def __init__(self, dest, sources, c0, gx=0.0, gy=0.0, gz=0.0, tdamp=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.c0 = c0
        self.tdamp = tdamp

        super(MomentumEquation, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, d_rho, d_p, d_V, d_au, d_av, d_aw,
             s_m, s_rho, s_p, s_V, DWIJ):

        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]
        p_i = d_p[d_idx]
        p_j = s_p[s_idx]

        pij = rhoj * p_i + rhoi * p_j
        pij /= (rhoj + rhoi)

        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # inverse mass of destination particle
        mi1 = 1.0/d_m[d_idx]

        tmp = -pij * mi1 * (Vi2 + Vj2)

        d_au[d_idx] += tmp * DWIJ[0]
        d_av[d_idx] += tmp * DWIJ[1]
        d_aw[d_idx] += tmp * DWIJ[2]

    def post_loop(self, d_idx, d_au, d_av, d_aw, t):
        damping_factor = 1.0
        if t < self.tdamp:
            damping_factor = 0.5 * (sin((-0.5 + t/self.tdamp)*M_PI) + 1.0)
        d_au[d_idx] += damping_factor*self.gx
        d_av[d_idx] += damping_factor*self.gy
        d_aw[d_idx] += damping_factor*self.gz


class EDACEquation(Equation):
    def __init__(self, dest, sources, cs, nu, rho0):
        self.cs = cs
        self.nu = nu
        self.rho0 = rho0

        super(EDACEquation, self).__init__(dest, sources)

    def initialize(self, d_idx, d_ap):
        d_ap[d_idx] = 0.0

    def loop(self, d_idx, d_m, d_rho, d_ap, d_p, d_V, s_idx, s_m, s_rho, s_p,
             s_V, DWIJ, VIJ, XIJ, R2IJ, EPS):

        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        etai = d_rho[d_idx]
        etaj = s_rho[s_idx]
        etaij = 2 * self.nu * (etai * etaj)/(etai + etaj)

        # This is the same as continuity acceleration times cs^2
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]
        vijdotdwij = DWIJ[0]*VIJ[0] + DWIJ[1]*VIJ[1] + DWIJ[2]*VIJ[2]
        d_ap[d_idx] += rhoi/rhoj*self.cs*self.cs*s_m[s_idx]*vijdotdwij

        # Viscous damping of pressure.
        xijdotdwij = DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2]
        tmp = 1.0/d_m[d_idx]*(Vi2 + Vj2)*etaij*xijdotdwij/(R2IJ + EPS)
        d_ap[d_idx] += tmp*(d_p[d_idx] - s_p[s_idx])


class MomentumEquationPressureGradient(Equation):

    r"""Momentum equation for internal flows with EDAC.

    This uses the basic formulation from the TVF scheme but modifies it to
    subtract an average pressure from Basa, Quinlan and Lastiwka correction
    from their 2009 paper.
    """

    def __init__(self, dest, sources, pb, gx=0., gy=0., gz=0.,
                 tdamp=0.0):

        r"""
        Parameters
        ----------
        pb : float
            background pressure
        gx : float
            Body force per unit mass along the x-axis
        gy : float
            Body force per unit mass along the y-axis
        gz : float
            Body force per unit mass along the z-axis
        tdamp : float
            damping time

        Notes
        -----
        This equation should have the destination as fluid and sources as
        fluid and boundary particles.

        This function also computes the contribution to the background
        pressure and accelerations due to a body force or gravity.

        The body forces are damped according to Eq. (13) in [Adami2012] to
        avoid instantaneous accelerations. By default, damping is neglected.
        """

        self.pb = pb
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.tdamp = tdamp
        super(MomentumEquationPressureGradient, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw, d_auhat, d_avhat, d_awhat):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

        d_auhat[d_idx] = 0.0
        d_avhat[d_idx] = 0.0
        d_awhat[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, d_rho, s_rho,
             d_au, d_av, d_aw, d_p, d_pavg, s_p,
             d_auhat, d_avhat, d_awhat, d_V, s_V, DWIJ):

        # averaged pressure Eq. (7)
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]
        pavg = d_pavg[d_idx]
        pi = d_p[d_idx]
        pj = s_p[s_idx]

        pij = rhoj * (pi - pavg) + rhoi * (pj - pavg)
        pij /= (rhoj + rhoi)

        # particle volumes
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # inverse mass of destination particle
        mi1 = 1.0/d_m[d_idx]

        # accelerations 1st term in Eq. (8)
        tmp = -pij * mi1 * (Vi2 + Vj2)

        d_au[d_idx] += tmp * DWIJ[0]
        d_av[d_idx] += tmp * DWIJ[1]
        d_aw[d_idx] += tmp * DWIJ[2]

        # contribution due to the background pressure Eq. (13)
        tmp = -self.pb * mi1 * (Vi2 + Vj2)

        d_auhat[d_idx] += tmp * DWIJ[0]
        d_avhat[d_idx] += tmp * DWIJ[1]
        d_awhat[d_idx] += tmp * DWIJ[2]

    def post_loop(self, d_idx, d_au, d_av, d_aw, t):
        # damped accelerations due to body or external force
        damping_factor = 1.0
        if t < self.tdamp:
            damping_factor = 0.5 * (sin((-0.5 + t/self.tdamp)*M_PI) + 1.0)

        d_au[d_idx] += self.gx * damping_factor
        d_av[d_idx] += self.gy * damping_factor
        d_aw[d_idx] += self.gz * damping_factor


class EDACTVFStep(IntegratorStep):

    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_p0, d_p):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_p0[d_idx] = d_p[d_idx]

    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_p0, d_p, d_au,
               d_av, d_auhat, d_avhat, d_awhat, d_uhat, d_vhat, d_what,
               d_aw, d_ap, dt=0.0):
        dtb2 = 0.5*dt
        d_u[d_idx] = d_u0[d_idx] + dtb2*d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2*d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2*d_aw[d_idx]

        d_uhat[d_idx] = d_u[d_idx] + dtb2*d_auhat[d_idx]
        d_vhat[d_idx] = d_v[d_idx] + dtb2*d_avhat[d_idx]
        d_what[d_idx] = d_w[d_idx] + dtb2*d_awhat[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dtb2 * d_uhat[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2 * d_vhat[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2 * d_what[d_idx]

        d_p[d_idx] = d_p0[d_idx] + dtb2 * d_ap[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_p0, d_p, d_au, d_av,
               d_aw, d_auhat, d_avhat, d_awhat, d_uhat, d_vhat, d_what,
               d_ap, dt=0.0):
        d_u[d_idx] = d_u0[d_idx] + dt*d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt*d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt*d_aw[d_idx]

        d_uhat[d_idx] = d_u[d_idx] + dt*d_auhat[d_idx]
        d_vhat[d_idx] = d_v[d_idx] + dt*d_avhat[d_idx]
        d_what[d_idx] = d_w[d_idx] + dt*d_awhat[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dt * d_uhat[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_vhat[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_what[d_idx]

        d_p[d_idx] = d_p0[d_idx] + dt * d_ap[d_idx]


class EDACScheme(Scheme):
    def __init__(self, fluids, solids, dim, c0, nu, rho0, pb=0.0,
                 gx=0.0, gy=0.0, gz=0.0, tdamp=0.0, eps=0.0, h=0.0,
                 edac_alpha=0.5, alpha=0.0, bql=True, clamp_p=False,
                 inlet_outlet_manager=None, inviscid_solids=None):
        """The EDAC scheme.

        Parameters
        ----------

        fluids : list(str)
            List of names of fluid particle arrays.
        solids : list(str)
            List of names of solid particle arrays.
        dim: int
            Dimensionality of the problem.
        c0 : float
            Speed of sound.
        nu : float
            Kinematic viscosity.
        rho0 : float
            Density of fluid.
        pb : float
            Background pressure value, if unset or zero, this uses an
            different formulation, else it uses the TVF with EDAC.
        gx, gy, gz : float
            Componenents of body acceleration (gravity, external forcing etc.)
        tdamp: float
            Time for which the acceleration should be damped.
        eps : float
            XSPH smoothing factor, defaults to zero.
        h : float
            Parameter h used for the particles -- used to calculate viscosity.
        edac_alpha : float
            Factor to use for viscosity.
        alpha : float
            Factor to use for artificial viscosity.
        bql : bool
            Use the Basa Quinlan Lastiwka correction.
        clamp_p : bool
            Clamp the boundary pressure to positive values.  This is only used
            for external flows.
        inlet_outlet_manager : InletOutletManager Instance
            Pass the manager if inlet outlet boundaries are present
        inviscid_solids : list
            list of inviscid solid array names
        """
        self.c0 = c0
        self.nu = nu
        self.rho0 = rho0
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.tdamp = tdamp
        self.dim = dim
        self.eps = eps
        self.fluids = fluids
        self.solids = solids
        self.pb = pb
        self.solver = None
        self.bql = bql
        self.clamp_p = clamp_p
        self.edac_alpha = edac_alpha
        self.alpha = alpha
        self.h = h
        self.inlet_outlet_manager = inlet_outlet_manager
        self.inviscid_solids = [] if inviscid_solids is None else\
            inviscid_solids
        self.attributes_changed()

    # Public protocol ###################################################
    def add_user_options(self, group):
        group.add_argument(
            "--alpha", action="store", type=float, dest="alpha",
            default=None,
            help="Alpha for the artificial viscosity."
        )
        group.add_argument(
            "--edac-alpha", action="store", type=float, dest="edac_alpha",
            default=None,
            help="Alpha for the EDAC scheme viscosity."
        )
        add_bool_argument(
            group, 'clamp-pressure', dest='clamp_p',
            help="Clamp pressure on boundaries to be non-negative.",
            default=None
        )
        add_bool_argument(
            group, 'use-bql', dest='bql',
            help="Use the Basa-Quinlan-Lastiwka correction.",
            default=None
        )
        group.add_argument(
            "--tdamp", action="store", type=float, dest="tdamp",
            default=None,
            help="Time for which the accelerations are damped."
        )

    def consume_user_options(self, options):
        vars = ['alpha', 'edac_alpha', 'clamp_p', 'bql', 'tdamp']
        data = dict((var, self._smart_getattr(options, var))
                    for var in vars)
        self.configure(**data)

    def attributes_changed(self):
        if self.pb is not None:
            self.use_tvf = abs(self.pb) > 1e-14
        if self.h is not None and self.c0 is not None:
            self.art_nu = self.edac_alpha*self.h*self.c0/8

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        This is to be called before `get_solver` is called.

        Parameters
        ----------

        dim : int
            Number of dimensions.
        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import QuinticSpline
        from pysph.sph.integrator import PECIntegrator
        if kernel is None:
            kernel = QuinticSpline(dim=self.dim)
        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        step_cls = EDACTVFStep if self.use_tvf else EDACStep
        cls = integrator_cls if integrator_cls is not None else PECIntegrator

        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = step_cls()

        iom = self.inlet_outlet_manager
        if iom is not None:
            iom_stepper = iom.get_stepper(self, cls, self.use_tvf)
            for name in iom_stepper:
                steppers[name] = iom_stepper[name]

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

        if iom is not None:
            iom.setup_iom(dim=self.dim, kernel=kernel)

    def get_equations(self):
        if self.use_tvf:
            return self._get_internal_flow_equations()
        else:
            return self._get_external_flow_equations()

    def get_solver(self):
        return self.solver

    def setup_properties(self, particles, clean=True):
        """Setup the particle arrays so they have the right set of properties
        for this scheme.

        Parameters
        ----------

        particles : list
            List of particle arrays.

        clean : bool
            If True, removes any unnecessary properties.
        """
        particle_arrays = dict([(p.name, p) for p in particles])
        TVF_FLUID_PROPS = set([
            'uhat', 'vhat', 'what', 'ap',
            'auhat', 'avhat', 'awhat', 'V',
            'p0', 'u0', 'v0', 'w0', 'x0', 'y0', 'z0',
            'pavg', 'nnbr'
        ])
        extra_props = TVF_FLUID_PROPS if self.use_tvf else EDAC_PROPS

        all_fluid_props = DEFAULT_PROPS.union(extra_props)
        iom = self.inlet_outlet_manager
        fluids_with_io = self.fluids
        if iom is not None:
            io_particles = iom.get_io_names(ghost=True)
            fluids_with_io = self.fluids + io_particles
        for fluid in fluids_with_io:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, all_fluid_props, clean)
            pa.set_output_arrays(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'p',
                                  'm', 'h', 'V'])
            if 'pavg' in pa.properties:
                pa.add_output_arrays(['pavg'])
            if iom is not None:
                iom.add_io_properties(pa, self)

        TVF_SOLID_PROPS = ['V', 'wij', 'ax', 'ay', 'az', 'uf', 'vf', 'wf',
                           'ug', 'vg', 'wg']
        if self.inviscid_solids:
            TVF_SOLID_PROPS += ['xn', 'yn', 'zn', 'uhat', 'vhat', 'what']
        extra_props = TVF_SOLID_PROPS if self.use_tvf else EDAC_SOLID_PROPS
        all_solid_props = DEFAULT_PROPS.union(extra_props)
        for solid in (self.solids+self.inviscid_solids):
            pa = particle_arrays[solid]
            self._ensure_properties(pa, all_solid_props, clean)
            pa.set_output_arrays(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'p',
                                  'm', 'h', 'V'])

    # Private protocol ###################################################
    def _get_edac_nu(self):
        if self.art_nu > 0:
            nu = self.art_nu
            print("Using artificial viscosity for EDAC with nu = %s" % nu)
        else:
            nu = self.nu
            print("Using real viscosity for EDAC with nu = %s" % self.nu)
        return nu

    def _get_internal_flow_equations(self):
        from pysph.sph.wc.transport_velocity import (
            VolumeSummation, SolidWallNoSlipBC, SummationDensity,
            MomentumEquationArtificialStress,
            MomentumEquationArtificialViscosity,
            MomentumEquationViscosity
        )
        edac_nu = self._get_edac_nu()

        iom = self.inlet_outlet_manager
        fluids_with_io = self.fluids
        all_solids = self.solids + self.inviscid_solids
        if iom is not None:
            fluids_with_io = self.fluids + iom.get_io_names()
        all = fluids_with_io + all_solids

        equations = []
        # inlet-outlet
        if iom is not None:
            io_eqns = iom.get_equations(self, self.use_tvf)
            for grp in io_eqns:
                equations.append(grp)

        group1 = []
        avg_p_group = []
        has_solids = len(all_solids) > 0
        for fluid in fluids_with_io:
            group1.append(SummationDensity(dest=fluid, sources=all))
            if self.bql:
                eq = ComputeAveragePressure(dest=fluid, sources=all)
                if has_solids:
                    avg_p_group.append(eq)
                else:
                    group1.append(eq)

        for solid in self.solids:
            group1.extend([
                SourceNumberDensity(dest=solid, sources=fluids_with_io),
                VolumeSummation(dest=solid, sources=all),
                SolidWallPressureBC(dest=solid, sources=fluids_with_io,
                                    gx=self.gx, gy=self.gy, gz=self.gz),
                SetWallVelocity(dest=solid, sources=fluids_with_io),
            ])
        for solid in self.inviscid_solids:
            group1.extend([
                SourceNumberDensity(dest=solid, sources=fluids_with_io),
                NoSlipVelocityExtrapolation(
                    dest=solid, sources=fluids_with_io),
                NoSlipAdvVelocityExtrapolation(
                    dest=solid, sources=fluids_with_io),
                VolumeSummation(dest=solid, sources=all),
                SolidWallPressureBC(dest=solid, sources=fluids_with_io,
                                    gx=self.gx, gy=self.gy, gz=self.gz)
                ])

        equations.append(Group(equations=group1, real=False))

        # Compute average pressure *after* the wall pressure is setup.
        if self.bql and has_solids:
            equations.append(Group(equations=avg_p_group, real=True))

        group2 = []
        for fluid in self.fluids:
            group2.append(
                MomentumEquationPressureGradient(
                    dest=fluid, sources=all, pb=self.pb,
                    gx=self.gx, gy=self.gy, gz=self.gz, tdamp=self.tdamp
                )
            )
            if self.alpha > 0.0:
                sources = fluids_with_io + self.solids
                group2.append(
                    MomentumEquationArtificialViscosity(
                        dest=fluid, sources=sources, alpha=self.alpha,
                        c0=self.c0
                    )
                )
            if self.nu > 0.0:
                group2.append(
                    MomentumEquationViscosity(
                        dest=fluid, sources=fluids_with_io, nu=self.nu
                    )
                )
            if len(self.solids) > 0 and self.nu > 0.0:
                group2.append(
                    SolidWallNoSlipBC(
                        dest=fluid, sources=self.solids, nu=self.nu
                    )
                )
            group2.extend([
                MomentumEquationArtificialStress(
                    dest=fluid, sources=fluids_with_io
                ),
                EDACEquation(
                    dest=fluid, sources=all, nu=edac_nu, cs=self.c0,
                    rho0=self.rho0
                ),
            ])
        equations.append(Group(equations=group2))

        # inlet-outlet
        if iom is not None:
            io_eqns = iom.get_equations_post_compute_acceleration()
            for grp in io_eqns:
                equations.append(grp)

        return equations

    def _get_external_flow_equations(self):

        from pysph.sph.basic_equations import XSPHCorrection
        from pysph.sph.wc.transport_velocity import (
            VolumeSummation, SolidWallNoSlipBC, SummationDensity,
            MomentumEquationArtificialViscosity,
            MomentumEquationViscosity
        )

        iom = self.inlet_outlet_manager
        fluids_with_io = self.fluids
        all_solids = self.solids + self.inviscid_solids
        if iom is not None:
            fluids_with_io = self.fluids + iom.get_io_names()
        all = fluids_with_io + all_solids

        edac_nu = self._get_edac_nu()
        equations = []
        # inlet-outlet
        if iom is not None:
            io_eqns = iom.get_equations(self, self.use_tvf)
            for grp in io_eqns:
                equations.append(grp)

        group1 = []
        for fluid in fluids_with_io:
            group1.append(SummationDensity(dest=fluid, sources=all))
        for solid in self.solids:
            group1.extend([
                SourceNumberDensity(dest=solid, sources=fluids_with_io),
                VolumeSummation(dest=solid, sources=all),
                SolidWallPressureBC(dest=solid, sources=fluids_with_io,
                                    gx=self.gx, gy=self.gy, gz=self.gz),
                SetWallVelocity(dest=solid, sources=fluids_with_io),
            ])
            if self.clamp_p:
                group1.append(
                    ClampWallPressure(dest=solid, sources=None)
                )

        for solid in self.inviscid_solids:
            group1.extend([
                SourceNumberDensity(dest=solid, sources=fluids_with_io),
                NoSlipVelocityExtrapolation(
                    dest=solid, sources=fluids_with_io),
                VolumeSummation(dest=solid, sources=all),
                SolidWallPressureBC(dest=solid, sources=fluids_with_io,
                                    gx=self.gx, gy=self.gy, gz=self.gz)
                ])

        equations.append(Group(equations=group1, real=False))

        group2 = []
        for fluid in self.fluids:
            group2.append(
                MomentumEquation(
                    dest=fluid, sources=all, gx=self.gx, gy=self.gy,
                    gz=self.gz, c0=self.c0, tdamp=self.tdamp
                )
            )
            if self.alpha > 0.0:
                sources = fluids_with_io + self.solids
                group2.append(
                    MomentumEquationArtificialViscosity(
                        dest=fluid, sources=sources, alpha=self.alpha,
                        c0=self.c0
                    )
                )
            if self.nu > 0.0:
                group2.append(
                    MomentumEquationViscosity(
                        dest=fluid, sources=fluids_with_io, nu=self.nu
                    )
                )
            if len(self.solids) > 0 and self.nu > 0.0:
                group2.append(
                    SolidWallNoSlipBC(
                        dest=fluid, sources=self.solids, nu=self.nu
                    )
                )
            group2.extend([
                EDACEquation(
                    dest=fluid, sources=all, nu=edac_nu, cs=self.c0,
                    rho0=self.rho0
                ),
                XSPHCorrection(dest=fluid, sources=[fluid],
                               eps=self.eps)
            ])
        equations.append(Group(equations=group2))

        # inlet-outlet
        if iom is not None:
            io_eqns = iom.get_equations_post_compute_acceleration()
            for grp in io_eqns:
                equations.append(grp)

        return equations



import unittest
import numpy as np
from math import sqrt
from pysph.examples.gas_dynamics import riemann_solver
import numpy.testing as npt


try:
    import scipy
except ImportError:
    scipy = None


class RiemannSolverTestCase(unittest.TestCase):

    def setUp(self):
        riemann_solver.set_gamma(1.4)

    def assert_error(self, given, expected, precision):
        return npt.assert_almost_equal(np.ravel(given), expected, precision)

    @unittest.skipIf(scipy is None, 'No scipy module, skipping this test')
    def test_compute_star_fsolve(self):
        # Sod Shock Tube
        x = riemann_solver.star_pu_fsolve(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.4),
            rho_r=0.125, u_r=0.0, p_r=0.1, c_r=sqrt(1.12)
            )
        self.assert_error(x, (0.30313, 0.92745), 5)
        # Sjogreen
        x = riemann_solver.star_pu_fsolve(
            rho_l=1.0, u_l=-2.0, p_l=0.4, c_l=sqrt(0.56),
            rho_r=1.0, u_r=2.0, p_r=0.4, c_r=sqrt(0.56)
            )
        self.assert_error(x, (0.00189, 0.00000), 5)
        # Blastwave
        x = riemann_solver.star_pu_fsolve(
            rho_l=1.0, u_l=0.0, p_l=1000.0, c_l=sqrt(1400),
            rho_r=1.0, u_r=0.0, p_r=0.01, c_r=sqrt(0.014)
            )
        self.assert_error(x, (460.894, 19.5975), 3)
        # Woodward and Collela
        x = riemann_solver.star_pu_fsolve(
            rho_l=1.0, u_l=0.0, p_l=0.01, c_l=sqrt(0.014),
            rho_r=1.0, u_r=0.0, p_r=100.0, c_r=sqrt(140)
            )
        self.assert_error(x, (46.0950, -6.19633), 4)
        # Sod Shock Tube
        riemann_solver.set_gamma(1.2)
        x = riemann_solver.star_pu_fsolve(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.2),
            rho_r=0.125, u_r=0.0, p_r=0.1, c_r=sqrt(0.96)
            )
        self.assert_error(x, (0.31274, 1.01132), 5)

    def test_compute_star_newton_raphson(self):
        # SodShock Tube
        x = riemann_solver.star_pu_newton_raphson(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.4),
            rho_r=0.125, u_r=0.0, p_r=0.1, c_r=sqrt(1.12)
            )
        self.assert_error(x, (0.30313, 0.92745), 5)

        # Sjogreen
        x = riemann_solver.star_pu_newton_raphson(
            rho_l=1.0, u_l=-2.0, p_l=0.4, c_l=sqrt(0.56),
            rho_r=1.0, u_r=2.0, p_r=0.4, c_r=sqrt(0.56)
            )
        self.assert_error(x, (0.00189, 0.00000), 5)

        # Blastwave
        x = riemann_solver.star_pu_newton_raphson(
            rho_l=1.0, u_l=0.0, p_l=1000.0, c_l=sqrt(1400),
            rho_r=1.0, u_r=0.0, p_r=0.01, c_r=sqrt(0.014)
            )
        self.assert_error(x, (460.894, 19.5975), 3)

        # Woodward and Collela
        x = riemann_solver.star_pu_newton_raphson(
            1.0, 0.0, 0.01, sqrt(0.014),
            1.0, 0.0, 100.0, sqrt(140)
            )
        self.assert_error(x, (46.0950, -6.19633), 4)

    def test_left_rarefaction(self):
        # SodShock Tube
        x = riemann_solver.left_rarefaction(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.4),
            p_star=0.30313, u_star=0.92745, s=-2.0
            )
        self.assert_error(x, (1.0, 0.0, 1.0), 10)
        x = riemann_solver.left_rarefaction(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.4),
            p_star=0.30313, u_star=0.92745, s=-1.6
            )
        self.assert_error(x, (1.0, 0.0, 1.0), 10)
        x = riemann_solver.left_rarefaction(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.4),
            p_star=0.30313, u_star=0.92745, s=0.0
            )
        self.assert_error(x, (0.42632, 0.92745, 0.30313), 5)

        # Blastwave
        x = riemann_solver.left_rarefaction(
            rho_l=1.0, u_l=0.0, p_l=1000.0, c_l=sqrt(1400),
            p_star=460.894, u_star=19.5975, s=-0.5 / 0.012
            )
        self.assert_error(x, (1.0, 0.0, 1000.0), 6)
        x = riemann_solver.left_rarefaction(
            rho_l=1.0, u_l=0.0, p_l=1000.0, c_l=sqrt(1400),
            p_star=460.894, u_star=19.5975, s=0.0 / 0.012
            )
        self.assert_error(x, (0.57506, 19.5975, 460.894), 5)

    def test_right_shock(self):
        # SodShock Tube
        x = riemann_solver.right_shock(
            rho_r=0.125, u_r=0.0, p_r=0.1, c_r=sqrt(1.4),
            p_star=0.30313, u_star=0.92745, s=2.0
            )
        self.assert_error(x, (0.125, 0.0, 0.1), 10)
        x = riemann_solver.right_shock(
            rho_r=0.125, u_r=0.0, p_r=0.1, c_r=sqrt(1.4),
            p_star=0.30313, u_star=0.92745, s=1.6
            )
        self.assert_error(x, (0.26557, 0.92745, 0.30313), 5)

        # Blastwave
        x = riemann_solver.right_shock(
            rho_r=1.0, u_r=0.0, p_r=0.01, c_r=sqrt(0.014),
            p_star=460.894, u_star=19.5975, s=0.5 / 0.012
            )
        self.assert_error(x, (1.0, 0.0, 0.01), 10)
        x = riemann_solver.right_shock(
            rho_r=1.0, u_r=0.0, p_r=0.01, c_r=sqrt(0.014),
            p_star=460.894, u_star=19.5975, s=0.2 / 0.012
            )
        self.assert_error(x, (5.99924, 19.5975, 460.894), 5)

    def test_right_rarefaction(self):
        # Sjogreen
        x = riemann_solver.right_rarefaction(
            rho_r=1.0, u_r=2.0, p_r=0.4, c_r=sqrt(0.56),
            p_star=0.00189, u_star=0.000, s=0.5 / 0.15
            )
        self.assert_error(x, (1.0, 2.0, 0.4), 10)
        x = riemann_solver.right_rarefaction(
            rho_r=1.0, u_r=2.0, p_r=0.4, c_r=sqrt(0.56),
            p_star=0.00189, u_star=0.000, s=0.0 / 0.15
            )
        self.assert_error(x, (0.02185, 0.0, 0.00189), 4)

    def test_left_shock(self):
        # Woodward and Collela
        x = riemann_solver.left_shock(
            rho_l=1.0, u_l=0.0, p_l=0.01, c_l=sqrt(0.014),
            p_star=46.095, u_star=-6.1963, s=-0.5 / 0.035
            )
        self.assert_error(x, (1.0, 0.0, 0.01), 10)
        x = riemann_solver.left_shock(
            rho_l=1.0, u_l=0.0, p_l=0.01, c_l=sqrt(0.014),
            p_star=46.095, u_star=-6.1963, s=-0.2 / 0.035
            )
        self.assert_error(x, (5.99242, -6.1963, 46.095), 5)

        # Test_case 5 (Toro)
        x = riemann_solver.left_shock(
            rho_l=5.99924, u_l=19.5975, p_l=460.894, c_l=sqrt(107.555),
            p_star=1691.64, u_star=8.68975, s=-0.5 / 0.035
            )
        self.assert_error(x, (5.99924, 19.5975, 460.894), 5)
        x = riemann_solver.left_shock(
            rho_l=5.99924, u_l=19.5975, p_l=460.894, c_l=sqrt(107.555),
            p_star=1691.64, u_star=8.68975, s=0.1 / 0.035
            )
        self.assert_error(x, (14.2823, 8.68975, 1691.64), 4)

    def test_left_contact(self):
        # Woodward and Collela
        x = riemann_solver.left_contact(
            rho_l=1.0, u_l=0.0, p_l=0.01, c_l=sqrt(0.014),
            p_star=46.095, u_star=-6.1963, s=-0.2 / 0.035
            )
        self.assert_error(x, (5.99242, -6.1963, 46.095), 5)

        # Test_case 5 (Toro)
        x = riemann_solver.left_contact(
            rho_l=5.99924, u_l=19.5975, p_l=460.894, c_l=sqrt(107.555),
            p_star=1691.64, u_star=8.68975, s=0.1 / 0.035
            )
        self.assert_error(x, (14.2823, 8.68975, 1691.64), 4)

        # SodShock Tube
        x = riemann_solver.left_contact(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.4),
            p_star=0.30313, u_star=0.92745, s=0.0
            )
        self.assert_error(x, (0.42632, 0.92745, 0.30313), 5)

        # Blastwave
        x = riemann_solver.left_contact(
            rho_l=1.0, u_l=0.0, p_l=1000.0, c_l=sqrt(1400),
            p_star=460.894, u_star=19.5975, s=0.0 / 0.012
            )
        self.assert_error(x, (0.57506, 19.5975, 460.894), 5)

    def test_right_contact(self):
        # Sjogreen
        x = riemann_solver.right_contact(
            rho_r=1.0, u_r=2.0, p_r=0.4, c_r=sqrt(0.56),
            p_star=0.00189, u_star=0.000, s=0.0 / 0.15
            )
        self.assert_error(x, (0.02185, 0.0, 0.00189), 4)

        # SodShock Tube
        x = riemann_solver.right_contact(
            rho_r=0.125, u_r=0.0, p_r=0.1, c_r=sqrt(1.4),
            p_star=0.30313, u_star=0.92745, s=1.6
            )
        self.assert_error(x, (0.26557, 0.92745, 0.30313), 5)

        # Blastwave
        x = riemann_solver.right_contact(
            rho_r=1.0, u_r=0.0, p_r=0.01, c_r=sqrt(0.014),
            p_star=460.894, u_star=19.5975, s=0.2 / 0.012
            )
        self.assert_error(x, (5.99924, 19.5975, 460.894), 5)

    def test_different_gamma(self):
        # SodShock Tube(gamma = 1.2)
        riemann_solver.set_gamma(1.2)
        x = riemann_solver.star_pu_newton_raphson(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.2),
            rho_r=0.125, u_r=0.0, p_r=0.1, c_r=sqrt(0.96)
            )
        self.assert_error(x, (0.31274, 1.01132), 5)
        x = riemann_solver.right_contact(
            rho_r=0.125, u_r=0.0, p_r=0.1, c_r=sqrt(1.2),
            p_star=0.31274, u_star=1.01132, s=0.15/0.1
            )
        self.assert_error(x, (0.31323, 1.01132, 0.31274), 5)
        x = riemann_solver.left_contact(
            rho_l=1.0, u_l=0.0, p_l=1.0, c_l=sqrt(1.2),
            p_star=0.31274, u_star=1.01132, s=-0.5/0.1
            )
        self.assert_error(x, (1.0, 0.0, 1.0), 10)

if __name__ == '__main__':
    unittest.main()



# Standard library imports.
import unittest

# Library imports.
import numpy as np
import pytest

# Local imports.
from pysph.base.utils import get_particle_array, get_particle_array_wcsph
from compyle.config import get_config
from pysph.sph.equation import Equation
from pysph.sph.acceleration_eval import AccelerationEval
from pysph.base.kernels import CubicSpline
from pysph.base.nnps import LinkedListNNPS
from pysph.sph.sph_compiler import SPHCompiler
from pysph.sph.integrator import (LeapFrogIntegrator, PECIntegrator,
                                  PEFRLIntegrator, EulerIntegrator)
from pysph.sph.integrator_step import (
    IntegratorStep, LeapFrogStep, PEFRLStep, TwoStageRigidBodyStep
)


class SHM(Equation):
    """Simple harmonic oscillator equation.
    """
    def initialize(self, d_idx, d_x, d_au):
        d_au[d_idx] = -d_x[d_idx]


class TestIntegrator(unittest.TestCase):
    def test_detection_of_missing_arrays_for_integrator(self):
        # Given.
        x = np.asarray([1.0])
        u = np.asarray([0.0])
        h = np.ones_like(x)
        pa = get_particle_array(name='fluid', x=x, u=u, h=h, m=h)
        arrays = [pa]

        # When
        integrator = LeapFrogIntegrator(fluid=LeapFrogStep())
        equations = [SHM(dest="fluid", sources=None)]
        kernel = CubicSpline(dim=1)
        a_eval = AccelerationEval(
            particle_arrays=arrays, equations=equations, kernel=kernel
        )
        comp = SPHCompiler(a_eval, integrator=integrator)

        # Then
        self.assertRaises(RuntimeError, comp.compile)

    def test_detect_missing_arrays_for_many_particle_arrays(self):
        # Given.
        x = np.asarray([1.0])
        u = np.asarray([0.0])
        h = np.ones_like(x)
        fluid = get_particle_array_wcsph(name='fluid', x=x, u=u, h=h, m=h)
        solid = get_particle_array(name='solid', x=x, u=u, h=h, m=h)
        arrays = [fluid, solid]

        # When
        integrator = PECIntegrator(
            fluid=TwoStageRigidBodyStep(), solid=TwoStageRigidBodyStep()
        )
        equations = [SHM(dest="fluid", sources=None)]
        kernel = CubicSpline(dim=1)
        a_eval = AccelerationEval(
            particle_arrays=arrays, equations=equations, kernel=kernel
        )
        comp = SPHCompiler(a_eval, integrator=integrator)

        # Then
        self.assertRaises(RuntimeError, comp.compile)


class TestIntegratorBase(unittest.TestCase):
    def setUp(self):
        x = np.asarray([1.0])
        u = np.asarray([0.0])
        h = np.ones_like(x)
        pa = get_particle_array(name='fluid', x=x, u=u, h=h, m=h)
        for prop in ('ax', 'ay', 'az', 'ae', 'arho', 'e'):
            pa.add_property(prop)
        self.pa = pa

    def _setup_integrator(self, equations, integrator):
        kernel = CubicSpline(dim=1)
        arrays = [self.pa]
        a_eval = AccelerationEval(
            particle_arrays=arrays, equations=equations, kernel=kernel
        )
        comp = SPHCompiler(a_eval, integrator=integrator)
        comp.compile()
        nnps = LinkedListNNPS(dim=kernel.dim, particles=arrays)
        a_eval.set_nnps(nnps)
        integrator.set_nnps(nnps)

    def _integrate(self, integrator, dt, tf, post_step_callback):
        """The post_step_callback is called after each step and is passed the
        current time.
        """
        t = 0.0
        while t < tf:
            integrator.step(t, dt)
            post_step_callback(t+dt)
            t += dt


class EulerStep(IntegratorStep):
    def stage1(self, d_idx, d_x, d_u, dt):
        d_x[d_idx] += dt*d_u[d_idx]


class TestIntegratorAdaptiveTimestep(TestIntegratorBase):
    def test_compute_timestep_without_adaptive(self):
        # Given.
        integrator = EulerIntegrator(fluid=EulerStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # When
        dt = integrator.compute_time_step(0.1, 0.5)

        # Then
        self.assertEqual(dt, None)

    def test_compute_timestep_with_dt_adapt(self):
        # Given.
        self.pa.extend(1)
        self.pa.align_particles()
        self.pa.add_property('dt_adapt')
        self.pa.dt_adapt[:] = [0.1, 0.2]

        integrator = EulerIntegrator(fluid=EulerStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # When
        dt = integrator.compute_time_step(0.1, 0.5)

        # Then
        self.assertEqual(dt, 0.1)

    def test_compute_timestep_with_dt_adapt_with_invalid_values(self):
        # Given.
        self.pa.extend(1)
        self.pa.align_particles()
        self.pa.add_property('dt_adapt')
        self.pa.dt_adapt[:] = [0.0, -2.0]

        integrator = EulerIntegrator(fluid=EulerStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # When
        dt = integrator.compute_time_step(0.1, 0.5)

        # Then
        self.assertEqual(dt, None)

    def test_compute_timestep_with_dt_adapt_trumps_dt_cfl(self):
        # Given.
        self.pa.extend(1)
        self.pa.align_particles()
        self.pa.add_property('dt_adapt')
        self.pa.add_property('dt_cfl')
        self.pa.h[:] = 1.0
        self.pa.dt_adapt[:] = [0.1, 0.2]
        self.pa.dt_cfl[:] = 1.0

        integrator = EulerIntegrator(fluid=EulerStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # When
        cfl = 0.5
        dt = integrator.compute_time_step(0.1, cfl)

        # Then
        self.assertEqual(dt, 0.1)

    def test_compute_timestep_with_dt_cfl(self):
        # Given.
        self.pa.extend(1)
        self.pa.align_particles()
        self.pa.add_property('dt_cfl')
        self.pa.h[:] = 1.0
        self.pa.dt_cfl[:] = [1.0, 2.0]

        integrator = EulerIntegrator(fluid=EulerStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # When
        cfl = 0.5
        dt = integrator.compute_time_step(0.1, cfl)

        # Then
        expect = cfl*1.0/(2.0)
        self.assertEqual(dt, expect)


class S1Step(IntegratorStep):

    def py_stage1(self, dest, t, dt):
        self.called_with1 = t, dt
        dest.u[:] = 1.0

    def stage1(self, d_idx, d_u, d_au, dt):
        d_u[d_idx] += d_au[d_idx] * dt * 0.5

    def stage2(self, d_idx, d_x, d_u, d_au, dt):
        d_u[d_idx] += 0.5*dt * d_au[d_idx]
        d_x[d_idx] += dt * d_u[d_idx]


class S12Step(IntegratorStep):

    def py_stage1(self, dest, t, dt):
        self.called_with1 = t, dt
        dest.u[:] = 1.0
        if dest.gpu:
            dest.gpu.push('u')

    def stage1(self, d_idx, d_u, d_au, dt):
        d_u[d_idx] += d_au[d_idx] * dt * 0.5

    def py_stage2(self, dest, t, dt):
        self.called_with2 = t, dt
        if dest.gpu:
            dest.gpu.pull('u')
        dest.u += 0.5
        if dest.gpu:
            dest.gpu.push('u')

    def stage2(self, d_idx, d_x, d_u, d_au, dt):
        d_u[d_idx] += 0.5*dt * d_au[d_idx]
        d_x[d_idx] += dt * d_u[d_idx]


class OnlyPyStep(IntegratorStep):

    def py_stage1(self, dest, t, dt):
        self.called_with1 = t, dt
        dest.x[:] = 0.0
        dest.u[:] = 1.0

    def py_stage2(self, dest, t, dt):
        self.called_with2 = t, dt
        dest.u += 0.5
        dest.x += 0.5


def my_helper(dt=0.0):
    return dt*2.0


class StepWithHelper(IntegratorStep):

    def _get_helpers_(self):
        return [my_helper]

    def stage1(self, d_idx, d_u, d_au, dt):
        d_u[d_idx] += d_au[d_idx] * my_helper(dt)


class TestLeapFrogIntegrator(TestIntegratorBase):
    def test_leapfrog(self):
        # Given.
        integrator = LeapFrogIntegrator(fluid=LeapFrogStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)
        tf = np.pi
        dt = 0.02*tf

        # When
        energy = []

        def callback(t):
            x, u = self.pa.x[0], self.pa.u[0]
            energy.append(0.5*(x*x + u*u))

        callback(0.0)
        self._integrate(integrator, dt, tf, callback)

        # Then
        energy = np.asarray(energy)
        self.assertAlmostEqual(np.max(np.abs(energy - 0.5)), 0.0, places=3)

    def test_integrator_calls_py_stage1(self):
        # Given.
        stepper = S1Step()
        integrator = LeapFrogIntegrator(fluid=stepper)
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)
        tf = 1.0
        dt = tf

        # When
        call_data = []

        def callback(t):
            call_data.append(t)

        self._integrate(integrator, dt, tf, callback)

        # Then
        self.assertEqual(len(call_data), 1)
        self.assertTrue(hasattr(stepper, 'called_with1'))
        self.assertEqual(stepper.called_with1, (0.0, dt))
        # These are not physically significant as the main purpose is to see
        # if the py_stage* methods are called.
        np.testing.assert_array_almost_equal(self.pa.x, [1.5])
        np.testing.assert_array_almost_equal(self.pa.u, [0.5])

    def test_integrator_calls_py_stage1_stage2(self):
        # Given.
        stepper = S12Step()
        integrator = LeapFrogIntegrator(fluid=stepper)
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)
        tf = 1.0
        dt = tf

        # When
        def callback(t):
            pass

        self._integrate(integrator, dt, tf, callback)

        # Then
        self.assertTrue(hasattr(stepper, 'called_with1'))
        self.assertEqual(stepper.called_with1, (0.0, dt))
        self.assertTrue(hasattr(stepper, 'called_with2'))
        self.assertEqual(stepper.called_with2, (0.5*dt, dt))
        # These are not physically significant as the main purpose is to see
        # if the py_stage* methods are called.
        np.testing.assert_array_almost_equal(self.pa.x, [2.0])
        np.testing.assert_array_almost_equal(self.pa.u, [1.0])

    def test_integrator_calls_only_py_when_no_stage(self):
        # Given.
        stepper = OnlyPyStep()
        integrator = LeapFrogIntegrator(fluid=stepper)
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)
        tf = 1.0
        dt = tf

        # When
        def callback(t):
            pass

        self._integrate(integrator, dt, tf, callback)

        # Then
        self.assertTrue(hasattr(stepper, 'called_with1'))
        self.assertEqual(stepper.called_with1, (0.0, dt))
        self.assertTrue(hasattr(stepper, 'called_with2'))
        self.assertEqual(stepper.called_with2, (0.5*dt, dt))
        np.testing.assert_array_almost_equal(self.pa.x, [0.5])
        np.testing.assert_array_almost_equal(self.pa.u, [1.5])

    def test_leapfrog_is_second_order(self):
        # Given.
        integrator = LeapFrogIntegrator(fluid=LeapFrogStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # Take a dt, find the error, halve dt, and see that error is drops as
        # desired.

        # When
        tf = np.pi
        dt = 0.02*tf
        energy = []

        def callback(t):
            x, u = self.pa.x[0], self.pa.u[0]
            energy.append(0.5*(x*x + u*u))

        callback(0.0)
        self._integrate(integrator, dt, tf, callback)
        energy = np.asarray(energy)
        err1 = np.max(np.abs(energy - 0.5))

        # When
        self.pa.x[0] = 1.0
        self.pa.u[0] = 0.0
        energy = []
        dt *= 0.5
        callback(0.0)
        self._integrate(integrator, dt, tf, callback)
        energy = np.asarray(energy)
        err2 = np.max(np.abs(energy - 0.5))

        # Then
        self.assertTrue(err2 < err1)
        self.assertAlmostEqual(err1/err2, 4.0, places=2)

    def test_helper_can_be_used_with_stepper(self):
        # Given.
        integrator = EulerIntegrator(fluid=StepWithHelper())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # When
        tf = 1.0
        dt = tf/2

        def callback(t):
            pass

        self._integrate(integrator, dt, tf, callback)

        # Then
        if self.pa.gpu is not None:
            self.pa.gpu.pull('u')
        u = self.pa.u
        self.assertEqual(u, -2.0*self.pa.x)


class TestPEFRLIntegrator(TestIntegratorBase):
    def test_pefrl(self):
        # Given.
        integrator = PEFRLIntegrator(fluid=PEFRLStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)
        tf = np.pi
        dt = 0.1*tf

        # When
        energy = []

        def callback(t):
            x, u = self.pa.x[0], self.pa.u[0]
            energy.append(0.5*(x*x + u*u))

        callback(0.0)
        self._integrate(integrator, dt, tf, callback)

        # Then
        energy = np.asarray(energy)
        self.assertAlmostEqual(np.max(np.abs(energy - 0.5)), 0.0, places=4)

    def test_pefrl_is_fourth_order(self):
        # Given.
        integrator = PEFRLIntegrator(fluid=PEFRLStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # Take a dt, find the error, halve dt, and see that error is drops as
        # desired.

        # When
        tf = np.pi
        dt = 0.1*tf
        energy = []

        def callback(t):
            x, u = self.pa.x[0], self.pa.u[0]
            energy.append(0.5*(x*x + u*u))

        callback(0.0)
        self._integrate(integrator, dt, tf, callback)
        energy = np.asarray(energy)
        err1 = np.max(np.abs(energy - 0.5))

        # When
        self.pa.x[0] = 1.0
        self.pa.u[0] = 0.0
        energy = []
        dt *= 0.5
        callback(0.0)
        self._integrate(integrator, dt, tf, callback)
        energy = np.asarray(energy)
        err2 = np.max(np.abs(energy - 0.5))

        # Then
        self.assertTrue(err2 < err1)
        self.assertTrue(err1/err2 > 16.0)


class TestLeapFrogIntegratorGPU(TestIntegratorBase):
    def _setup_integrator(self, equations, integrator):
        pytest.importorskip('pyopencl')
        pytest.importorskip('pysph.base.gpu_nnps')
        kernel = CubicSpline(dim=1)
        arrays = [self.pa]
        from pysph.base.gpu_nnps import BruteForceNNPS as GPUNNPS
        a_eval = AccelerationEval(
             particle_arrays=arrays, equations=equations, kernel=kernel,
             backend='opencl'
        )
        comp = SPHCompiler(a_eval, integrator=integrator)
        comp.compile()
        nnps = GPUNNPS(dim=kernel.dim, particles=arrays, cache=True,
                       backend='opencl')
        nnps.update()
        a_eval.set_nnps(nnps)
        integrator.set_nnps(nnps)

    def test_leapfrog(self):
        # Given.
        integrator = LeapFrogIntegrator(fluid=LeapFrogStep())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)
        tf = np.pi/5
        dt = 0.05*tf

        # When
        energy = []

        def callback(t):
            self.pa.gpu.pull('x', 'u')
            x, u = self.pa.x[0], self.pa.u[0]
            energy.append(0.5*(x*x + u*u))

        callback(0.0)
        self._integrate(integrator, dt, tf, callback)

        # Then
        energy = np.asarray(energy)
        self.assertAlmostEqual(np.max(np.abs(energy - 0.5)), 0.0, places=3)

    def test_py_stage_is_called_on_gpu(self):
        # Given.
        stepper = S12Step()
        integrator = LeapFrogIntegrator(fluid=stepper)
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)
        dt = 1.0
        tf = dt

        # When
        def callback(t):
            pass

        self._integrate(integrator, dt, tf, callback)
        self.pa.gpu.pull('x', 'u')

        # Then
        self.assertTrue(hasattr(stepper, 'called_with1'))
        self.assertEqual(stepper.called_with1, (0.0, dt))
        self.assertTrue(hasattr(stepper, 'called_with2'))
        self.assertEqual(stepper.called_with2, (0.5*dt, dt))
        # These are not physically significant as the main purpose is to see
        # if the py_stage* methods are called.
        np.testing.assert_array_almost_equal(self.pa.x, [2.0])
        np.testing.assert_array_almost_equal(self.pa.u, [1.0])

    def test_leapfrog_with_double(self):
        orig = get_config().use_double

        def _cleanup():
            get_config().use_double = orig
        get_config().use_double = True
        self.addCleanup(_cleanup)
        self.test_leapfrog()

    def test_helper_can_be_used_with_stepper_on_gpu(self):
        # Given.
        integrator = EulerIntegrator(fluid=StepWithHelper())
        equations = [SHM(dest="fluid", sources=None)]
        self._setup_integrator(equations=equations, integrator=integrator)

        # When
        tf = 1.0
        dt = tf/2

        def callback(t):
            pass

        self._integrate(integrator, dt, tf, callback)

        # Then
        if self.pa.gpu is not None:
            self.pa.gpu.pull('u')
        u = self.pa.u
        self.assertEqual(u, -2.0*self.pa.x)


class TestLeapFrogIntegratorCUDA(TestLeapFrogIntegratorGPU):
    def _setup_integrator(self, equations, integrator):
        pytest.importorskip('pycuda')
        pytest.importorskip('pysph.base.gpu_nnps')
        kernel = CubicSpline(dim=1)
        arrays = [self.pa]
        from pysph.base.gpu_nnps import BruteForceNNPS as GPUNNPS
        a_eval = AccelerationEval(
             particle_arrays=arrays, equations=equations, kernel=kernel,
             backend='cuda'
        )
        comp = SPHCompiler(a_eval, integrator=integrator)
        comp.compile()
        nnps = GPUNNPS(dim=kernel.dim, particles=arrays, cache=True,
                       backend='cuda')
        nnps.update()
        a_eval.set_nnps(nnps)
        integrator.set_nnps(nnps)


if __name__ == '__main__':
    unittest.main()



# cython: language_level=3, embedsignature=True
# distutils: language=c++
# numpy
cimport numpy as np
cimport cython

from libcpp.map cimport map
from libcpp.vector cimport vector

import pyopencl as cl
import pyopencl.array

# PyZoltan CArrays
from cyarray.carray cimport UIntArray, IntArray, DoubleArray, LongArray

# local imports
from .particle_array cimport ParticleArray
from .point cimport *

from pysph.base.nnps_base cimport *

cdef extern from 'math.h':
    int abs(int) nogil
    double ceil(double) nogil
    double floor(double) nogil
    double fabs(double) nogil
    double fmax(double, double) nogil
    double fmin(double, double) nogil

cdef extern from 'limits.h':
    cdef unsigned int UINT_MAX
    cdef int INT_MAX


cdef class GPUNeighborCache:
    cdef object backend
    cdef int _dst_index
    cdef int _src_index
    cdef int _narrays
    cdef list _particles

    cdef bint _cached
    cdef public bint _copied_to_cpu
    cdef GPUNNPS _nnps

    cdef public object _neighbors_gpu
    cdef public object _nbr_lengths_gpu
    cdef public object _start_idx_gpu

    cdef object _get_start_indices

    cdef public np.ndarray _neighbors_cpu
    cdef public np.ndarray _nbr_lengths
    cdef public np.ndarray _start_idx

    cdef unsigned int* _neighbors_cpu_ptr
    cdef unsigned int* _nbr_lengths_ptr
    cdef unsigned int* _start_idx_ptr

    cdef void copy_to_cpu(self)
    cdef void get_neighbors_raw(self, size_t d_idx, UIntArray nbrs)
    cdef void get_neighbors_raw_gpu(self)
    cpdef update(self)

    cdef void _find_neighbors(self)
    cpdef get_neighbors(self, int src_index, size_t d_idx, UIntArray nbrs)
    cpdef get_neighbors_gpu(self)

cdef class GPUNNPS(NNPSBase):

    cdef public object backend
    cdef public object queue

    cdef public double radius_scale2
    cdef public GPUNeighborCache current_cache  # The current cache
    cdef public bint sort_gids        # Sort neighbors by their gids.
    cdef public bint use_double
    cdef public dtype
    cdef public dtype_max
    cdef public double _last_domain_size # last size of domain.

    cdef public np.ndarray xmin
    cdef public np.ndarray xmax

    cpdef get_nearest_particles(self, int src_index, int dst_index,
            size_t d_idx, UIntArray nbrs)

    cpdef get_nearest_particles_gpu(self, int src_index, int dst_index)

    cpdef spatially_order_particles(self, int pa_index)

    cdef void get_nearest_neighbors(self, size_t d_idx, UIntArray nbrs)

    cdef void find_neighbor_lengths(self, nbr_lengths)

    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices)

    cdef _compute_bounds(self)

    cpdef update(self)

    cpdef _bin(self, int pa_index)

    cpdef _refresh(self)

cdef class BruteForceNNPS(GPUNNPS):
    cdef NNPSParticleArrayWrapper src, dst # Current source and destination.
    cdef str preamble

    cpdef set_context(self, int src_index, int dst_index)

    cdef void find_neighbor_lengths(self, nbr_lengths)

    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices)

    cpdef _refresh(self)



from pysph.base.gpu_helper_kernels import *
from compyle.api import declare
from compyle.template import Template
from compyle.low_level import atomic_inc


@annotate
def fill_pids(i, x, y, z, cell_size, xmin, ymin, zmin, keys, pids):
    c = declare('matrix(3, "int")')
    find_cell_id(
        x[i] - xmin,
        y[i] - ymin,
        z[i] - zmin,
        cell_size, c
    )
    key = declare('ulong')
    key = interleave3(c[0], c[1], c[2])
    keys[i] = key
    pids[i] = i


@annotate
def count_cells(i, keys, cell_size, cell_lengths,
                sort_offsets):
    key = keys[i]
    idx = atomic_inc(cell_lengths[key])
    sort_offsets[i] = idx


@annotate
def sort_indices(i, keys, sort_offsets, start_indices, sorted_indices,
                 sorted_keys):
    key = keys[i]
    offset = sort_offsets[i]
    start_idx = start_indices[key]
    sorted_indices[start_idx + offset] = i
    sorted_keys[start_idx + offset] = key


@annotate
def input_start_indices(i, counts):
    return 0 if i == 0 else counts[i - 1]


@annotate
def output_start_indices(i, item, indices):
    indices[i] = item


@annotate
def inp_fill_unique_cids(i, keys, cids):
    return 1 if i != 0 and keys[i] != keys[i - 1] else 0


@annotate
def out_fill_unique_cids(i, item, cids):
    cids[i] = item


@annotate
def map_cid_to_idx(i, x, y, z, num_particles, cell_size, xmin, ymin, zmin,
                   pids, keys, cids, cid_to_idx):
    cid = cids[i]

    if i != 0 and cids[i] == cids[i - 1]:
        return

    c = declare('matrix(3, "int")')

    pid = pids[i]

    find_cell_id(
        x[pid] - xmin,
        y[pid] - ymin,
        z[pid] - zmin,
        cell_size, c
    )

    nbr_boxes = declare('matrix(27, "ulong")')

    nbr_boxes_length = neighbor_boxes(c[0], c[1], c[2], nbr_boxes)

    for j in range(nbr_boxes_length):
        key = nbr_boxes[j]
        idx = find_idx(keys, num_particles, key)
        cid_to_idx[27 * cid + j] = idx


@annotate
def map_dst_to_src(i, dst_to_src, cids_dst, cid_to_idx_dst,
                   keys_dst, keys_src, cids_src, num_particles_src, max_cid_src):
    idx = cid_to_idx_dst[27 * i]
    key = keys_dst[idx]
    idx_src = find_idx(keys_src, num_particles_src, key)
    dst_to_src[i] = atomic_inc(
        max_cid_src[0]) if (
        idx_src == -
        1) else cids_src[idx_src]


@annotate
def fill_overflow_map(i, dst_to_src, cid_to_idx_dst, x, y, z,
                      num_particles_src, cell_size, xmin, ymin, zmin, keys_src,
                      pids_dst, overflow_cid_to_idx, max_cid_src):
    cid = dst_to_src[i]
    # i is the cid in dst

    if cid < max_cid_src:
        return

    idx = cid_to_idx_dst[27 * i]

    pid = pids_dst[idx]

    c = declare('matrix(3, "int")')

    find_cell_id(
        x[pid] - xmin,
        y[pid] - ymin,
        z[pid] - zmin,
        cell_size, c
    )

    nbr_boxes = declare('matrix(27, "ulong")')

    nbr_boxes_length = neighbor_boxes(c[0], c[1], c[2], nbr_boxes)

    start_idx = cid - max_cid_src

    for j in range(nbr_boxes_length):
        key = nbr_boxes[j]
        idx = find_idx(keys_src, num_particles_src, key)
        overflow_cid_to_idx[27 * start_idx + j] = idx


class ZOrderNNPSKernel(Template):
    def __init__(self, name, dst_src=False, z_order_length=False,
                 z_order_nbrs=False):
        super(ZOrderNNPSKernel, self).__init__(name=name)
        self.z_order_length = z_order_length
        self.z_order_nbrs = z_order_nbrs
        assert self.z_order_nbrs != self.z_order_length
        self.dst_src = dst_src

    def template(self, i, d_x, d_y, d_z, d_h, s_x, s_y, s_z, s_h, xmin, ymin,
                 zmin, num_particles, keys, pids_dst, pids_src, max_cid_src,
                 cids, cid_to_idx, overflow_cid_to_idx, dst_to_src,
                 radius_scale2, cell_size):
        '''
        q = declare('matrix(4)')

        qid = pids_dst[i]

        q[0] = d_x[qid]
        q[1] = d_y[qid]
        q[2] = d_z[qid]
        q[3] = d_h[qid]

        cid = cids[i]
        nbr_boxes = declare('GLOBAL_MEM int*')
        nbr_boxes = cid_to_idx
        h_i = radius_scale2*q[3]*q[3]

        % if obj.dst_src:
        cid = dst_to_src[cid]
        start_id_nbr_boxes = 27*cid
        if cid >= max_cid_src:
            start_id_nbr_boxes = 27*(cid - max_cid_src)
            nbr_boxes = overflow_cid_to_idx
        % else:
        start_id_nbr_boxes = 27*cid
        % endif

        % if obj.z_order_length:
        length = 0
        % elif obj.z_order_nbrs:
        start_idx = start_indicies[qid]
        curr_idx = 0
        % endif

        s = declare('matrix(4)')
        j = declare('int')

        for j in range(27):
            idx = nbr_boxes[start_id_nbr_boxes + j]
            if idx == -1:
                continue
            key = keys[idx]

            while (idx < num_particles and keys[idx] == key):
                pid = pids_src[idx]
                s[0] = s_x[pid]
                s[1] = s_y[pid]
                s[2] = s_z[pid]
                s[3] = s_h[pid]

                h_j = radius_scale2 * s[3] * s[3]

                % if obj.z_order_nbrs:
                dist = norm2(q[0] - s[0], q[1] - s[1], q[2] - s[2])
                if dist < h_i or dist < h_j:
                    nbrs[start_idx + curr_idx] = pid
                    curr_idx += 1
                % else:
                dist = norm2(q[0] - s[0], q[1] - s[1], q[2] - s[2])
                if dist < h_i or dist < h_j:
                    length += 1
                %endif
                idx += 1

        % if obj.z_order_length:
        nbr_lengths[qid] = length
        % endif
        '''


class ZOrderLengthKernel(ZOrderNNPSKernel):
    def __init__(self, name, dst_src):
        super(ZOrderLengthKernel, self).__init__(
            name, dst_src, z_order_length=True,
        )

    def extra_args(self):
        return ['nbr_lengths'], {}


class ZOrderNbrsKernel(ZOrderNNPSKernel):
    def __init__(self, name, dst_src):
        super(ZOrderNbrsKernel, self).__init__(
            name, dst_src, z_order_nbrs=True
        )

    def extra_args(self):
        return ['start_indicies', 'nbrs'], {}



"""Boundary equations for Gas-dynamics"""

from pysph.sph.equation import Equation

class WallBoundary(Equation):
    def initialize(self, d_idx, d_p,  d_rho, d_e, d_m, d_cs, d_div, d_h,
            d_htmp, d_h0, d_u, d_v, d_w,  d_wij):
        d_p[d_idx] = 0.0
        d_u[d_idx] = 0.0
        d_v[d_idx] = 0.0
        d_w[d_idx] = 0.0
        d_m[d_idx] = 0.0
        d_rho[d_idx] = 0.0
        d_e[d_idx] = 0.0
        d_cs[d_idx] = 0.0
        d_div[d_idx] = 0.0
        d_wij[d_idx] = 0.0
        d_h[d_idx]  = d_h0[d_idx]
        d_htmp[d_idx]  = 0.0

    def loop(self, d_idx, s_idx, d_p, d_rho, d_e, d_m, d_cs, d_div, d_h, d_u,
            d_v, d_w, d_wij, d_htmp, s_p, s_rho, s_e, s_m, s_cs, s_h, s_div,
            s_u, s_v, s_w,  WI):
        d_wij[d_idx] += WI
        d_p[d_idx] += s_p[s_idx]*WI
        d_u[d_idx] -= s_u[s_idx]*WI
        d_v[d_idx] -= s_v[s_idx]*WI
        d_w[d_idx] -= s_w[s_idx]*WI
        d_m[d_idx] += s_m[s_idx]*WI
        d_rho[d_idx] +=  s_rho[s_idx]*WI
        d_e[d_idx] += s_e[s_idx]*WI
        d_cs[d_idx] += s_cs[s_idx]*WI
        d_div[d_idx] += s_div[s_idx]*WI
        d_htmp[d_idx]  += s_h[s_idx]*WI

    def post_loop(self, d_idx, d_p, d_rho, d_e, d_m, d_cs, d_div, d_h, d_u,
             d_v, d_w, d_wij, d_htmp):
        if (d_wij[d_idx]>1e-30):
            d_p[d_idx] = d_p[d_idx]/d_wij[d_idx]
            d_u[d_idx] = d_u[d_idx]/d_wij[d_idx]
            d_v[d_idx] = d_v[d_idx]/d_wij[d_idx]
            d_w[d_idx] = d_w[d_idx]/d_wij[d_idx]
            d_m[d_idx] = d_m[d_idx]/d_wij[d_idx]
            d_rho[d_idx] = d_rho[d_idx]/d_wij[d_idx]
            d_e[d_idx] =  d_e[d_idx]/d_wij[d_idx]
            d_cs[d_idx] =  d_cs[d_idx]/d_wij[d_idx]
            d_div[d_idx] =  d_div[d_idx]/d_wij[d_idx]
            d_h[d_idx] =  d_htmp[d_idx] /d_wij[d_idx]



#cython: embedsignature=True

from .nnps_base cimport *
from .octree cimport Octree, CompressedOctree, cOctreeNode

from libcpp.vector cimport vector
from libc.stdlib cimport malloc, free

cimport cython
from cython.operator cimport dereference as deref, preincrement as inc

IF UNAME_SYSNAME == "Windows":
    cdef inline double fmin(double x, double y) nogil:
        return x if x < y else y
    cdef inline double fmax(double x, double y) nogil:
        return x if x > y else y


#############################################################################
cdef class OctreeNNPS(NNPS):
    """Nearest neighbor search using Octree.
    """
    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int leaf_max_particles = 10, bint test_parallel = False):
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        cdef int i
        self.tree = [Octree(leaf_max_particles) for i in range(self.narrays)]

        self.radius_scale2 = radius_scale*radius_scale

        self.src_index = 0
        self.dst_index = 0
        self.leaf_max_particles = leaf_max_particles
        self.current_tree = NULL
        self.current_pids = NULL

        for i in range(self.narrays):
            self.tree[i].test_parallel = test_parallel

        self.sort_gids = sort_gids
        self.domain.update()
        self.update()


    #### Public protocol ################################################

    cpdef get_depth(self, int pa_index):
        return (<Octree>self.tree[pa_index]).depth

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """

        NNPS.set_context(self, src_index, dst_index)
        self.current_tree = (<Octree>self.tree[src_index]).root
        self.current_pids = (<Octree>self.tree[src_index]).pids

        self.dst = <NNPSParticleArrayWrapper> self.pa_wrappers[dst_index]
        self.src = <NNPSParticleArrayWrapper> self.pa_wrappers[src_index]

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        """Low level, high-performance non-gil method to find neighbors.
        This requires that `set_context()` be called beforehand.  This method
        does not reset the neighbors array before it appends the
        neighbors to it.

        """
        cdef double* dst_x_ptr = self.dst.x.data
        cdef double* dst_y_ptr = self.dst.y.data
        cdef double* dst_z_ptr = self.dst.z.data
        cdef double* dst_h_ptr = self.dst.h.data

        cdef double* src_x_ptr = self.src.x.data
        cdef double* src_y_ptr = self.src.y.data
        cdef double* src_z_ptr = self.src.z.data
        cdef double* src_h_ptr = self.src.h.data

        cdef double x = dst_x_ptr[d_idx]
        cdef double y = dst_y_ptr[d_idx]
        cdef double z = dst_z_ptr[d_idx]
        cdef double h = dst_h_ptr[d_idx]

        cdef u_int* s_gid = self.src.gid.data
        cdef int orig_length = nbrs.length

        self._get_neighbors(x, y, z, h, src_x_ptr, src_y_ptr, src_z_ptr,
                src_h_ptr, nbrs, self.current_tree)

        if self.sort_gids:
            self._sort_neighbors(
                &nbrs.data[orig_length], nbrs.length - orig_length, s_gid
            )


    #### Private protocol ################################################

    @cython.cdivision(True)
    cdef void _get_neighbors(self, double q_x, double q_y, double q_z, double q_h,
            double* src_x_ptr, double* src_y_ptr, double* src_z_ptr, double* src_h_ptr,
            UIntArray nbrs, cOctreeNode* node) nogil:
        """Find neighbors recursively"""
        cdef double x_centre = node.xmin[0] + node.length/2
        cdef double y_centre = node.xmin[1] + node.length/2
        cdef double z_centre = node.xmin[2] + node.length/2

        cdef u_int i, j, k
        cdef double hi2 = self.radius_scale2*q_h*q_h
        cdef double hj2 = 0
        cdef double xij2 = 0

        cdef double eff_radius = 0.5*(node.length) + \
                fmax(self.radius_scale*q_h, self.radius_scale*node.hmax)

        if  fabs(x_centre - q_x) >= eff_radius or \
            fabs(y_centre - q_y) >= eff_radius or \
            fabs(z_centre - q_z) >= eff_radius:
            return

        if node.is_leaf:
            for i from 0<=i<node.num_particles:
                k = self.current_pids[node.start_index + i]
                hj2 = self.radius_scale2*src_h_ptr[k]*src_h_ptr[k]
                xij2 = norm2(
                        src_x_ptr[k] - q_x,
                        src_y_ptr[k] - q_y,
                        src_z_ptr[k] - q_z
                        )
                if (xij2 < hi2) or (xij2 < hj2):
                    nbrs.c_append(k)
            return

        for i from 0<=i<8:
            if node.children[i] == NULL:
                continue
            self._get_neighbors(q_x, q_y, q_z, q_h,
                    src_x_ptr, src_y_ptr, src_z_ptr, src_h_ptr,
                    nbrs, node.children[i])

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices):
        indices.reset()
        cdef int num_particles = (<Octree>self.tree[pa_index]).num_particles
        cdef u_int* current_pids = (<Octree>self.tree[pa_index]).pids

        cdef int i
        for i from 0<=i<num_particles:
            indices.c_append(<long>current_pids[i])

    cpdef _refresh(self):
        cdef int i
        for i from 0<=i<self.narrays:
            (<Octree>self.tree[i]).c_build_tree(self.pa_wrappers[i], self.tree[i].test_parallel)
        self.current_tree = (<Octree>self.tree[self.src_index]).root
        self.current_pids = (<Octree>self.tree[self.src_index]).pids

    cpdef _bin(self, int pa_index, UIntArray indices):
        pass

#############################################################################
cdef class CompressedOctreeNNPS(OctreeNNPS):
    """Nearest neighbor search using Compressed Octree.
    """
    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int leaf_max_particles = 10,
            bint test_parallel = False):
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        cdef int i
        self.tree = [CompressedOctree(leaf_max_particles) for i in range(self.narrays)]

        self.radius_scale2 = radius_scale*radius_scale

        self.src_index = 0
        self.dst_index = 0
        self.leaf_max_particles = leaf_max_particles
        self.current_tree = NULL
        self.current_pids = NULL

        for i in range(self.narrays):
            self.tree[i].test_parallel = test_parallel

        self.sort_gids = sort_gids
        self.domain.update()
        self.update()


    #### Public protocol ################################################

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """

        NNPS.set_context(self, src_index, dst_index)
        self.current_tree = (<CompressedOctree>self.tree[src_index]).root
        self.current_pids = (<CompressedOctree>self.tree[src_index]).pids

        self.dst = <NNPSParticleArrayWrapper> self.pa_wrappers[dst_index]
        self.src = <NNPSParticleArrayWrapper> self.pa_wrappers[src_index]


    #### Private protocol ################################################

    cpdef _refresh(self):
        cdef int i
        for i from 0<=i<self.narrays:
            (<CompressedOctree>self.tree[i]).c_build_tree(self.pa_wrappers[i],
                                                          self.tree[i].test_parallel)
        self.current_tree = (<CompressedOctree>self.tree[self.src_index]).root
        self.current_pids = (<CompressedOctree>self.tree[self.src_index]).pids

    cpdef _bin(self, int pa_index, UIntArray indices):
        pass



''' convert pysph .npz output to vtk file format '''
from __future__ import print_function
import os
import re

from enthought.tvtk.api import tvtk, write_data
from numpy import array, c_, ravel, load, zeros_like


def write_vtk(data, filename, scalars=None, vectors={'V':('u','v','w')}, tensors={},
              coords=('x','y','z'), dims=None, **kwargs):
    ''' write data in to vtk file

    Parameters
    ----------
    data : dict
        mapping of variable name to their numpy array
    filename : str
        the file to write to (can be any recognized vtk extension)
        if extension is missing .vts extension is appended
    scalars : list
        list of arrays to write as scalars (defaults to data.keys())
    vectors : dict
        mapping of vector name to vector component names to take from data
    tensors : dict
        mapping of tensor name to tensor component names to take from data
    coords : list
        the name of coordinate data arrays (default=('x','y','z'))
    dims : 3 tuple
        the size along the dimensions for (None means x.shape)
    **kwargs : extra arguments for the file writer
        example file_type=binary/ascii

    '''
    x = data[coords[0]]
    y = data.get(coords[1], zeros_like(x))
    z = data.get(coords[2], zeros_like(x))

    if dims is None:
        dims = array([1,1,1])
        dims[:x.ndim] = x.shape
    else:
        dims = array(dims)

    sg = tvtk.StructuredGrid(points=c_[x.flat,y.flat,z.flat],dimensions=array(dims))
    pd = tvtk.PointData()

    if scalars is None:
        scalars = [i for i in data.keys() if i not in coords]
    for v in scalars:
        pd.scalars = ravel(data[v])
        pd.scalars.name = v
        sg.point_data.add_array(pd.scalars)

    for vec,vec_vars in vectors.items():
        u,v,w = [data[i] for i in vec_vars]
        pd.vectors = c_[ravel(u),ravel(v),ravel(w)]
        pd.vectors.name = vec
        sg.point_data.add_array(pd.vectors)

    for ten,ten_vars in tensors.items():
        vars = [data[i] for i in ten_vars]
        tensors = c_[[ravel(i) for i in vars]].T
        pd.tensors = tensors
        pd.tensors.name = ten
        sg.point_data.add_array(pd.tensors)

    write_data(sg, filename, **kwargs)


def detect_vectors_tensors(keys):
    ''' detect the vectors and tensors from given array names

    Vectors are identified as the arrays with common prefix followed by
    0,1 and 2 in their names
    Tensors are identified as the arrays with common prefix followed by
    two character codes representing ij indices
        (00,01,02,11,12,22) for a symmetric tensor
        (00,01,02,10,11,12,20,21,22) for a tensor
    Arrays not belonging to vectors or tensors are returned as scalars

    Returns scalars,vectors,tensors in a format suitable to be used as arguments
    for :py:func:`write_vtk`

    '''
    d = {}
    for k in keys:
        d[len(k)] = d.get(len(k), [])
        d[len(k)].append(k)

    scalars = []
    vectors = {}
    tensors = {}

    for n,l in d.items():
        if n<2:
            continue
        l.sort()

        idx = -1
        while idx<len(l)-1:
            idx += 1
            k = l[idx]

            # check if last char is 0
            if k[-1] == '0':

                # check for tensor
                if k[-2] == '0':

                    # check for 9 tensor
                    ten = []
                    for i in range(3):
                        for j in range(3):
                            ten.append(k[:-2]+str(j)+str(i))
                    ten.sort()

                    if l[idx:idx+9] == ten:
                        tensors[k[:-2]] = ten
                        idx += 8
                        continue

                    # check for symm 6 tensor
                    ten2 = []
                    for i in range(3):
                        for j in range(i+1):
                            ten2.append(k[:-2]+str(j)+str(i))
                    ten2.sort()

                    if l[idx:idx+6] == ten2:
                        ten = []
                        for i in range(3):
                            for j in range(3):
                                ten.append(k[:-2]+str(min(i,j))+str(max(i,j)))
                        tensors[k[:-2]] = ten
                        idx += 5
                        continue

                # check for vector
                vec = []
                for i in range(3):
                    vec.append(k[:-1] + str(i))
                if l[idx:idx+3] == vec:
                    vectors[k[:-1]] = vec
                    idx += 2
                    continue

            scalars.append(k)

    return scalars, vectors, tensors



def get_output_details(path):
    solvers = {}
    if not os.path.isdir(path):
        path = os.path.dirname(path)
    files = os.listdir(path)
    files.sort()

    pat = re.compile(r'(?P<solver>.+)_(?P<rank>\d+)_(?P<entity>.+)_(?P<time>[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?).npz')
    matches = [(f,pat.match(f)) for f in files]

    files = []
    for filename,match in matches:
        if match is None:
            continue
        files.append(filename)
        groups = match.groupdict()
        solvername = groups['solver']
        solver = solvers.get(solvername)
        if solver is None:
            solver = [set([]),set([]),set([])]
            solvers[solvername] = solver
        solver[0].add(groups['rank'])
        solver[1].add(groups['entity'])
        solver[2].add(groups['time'])
    # {solver:(entities,procs,times)}
    return solvers


def pysph_to_vtk(path, merge_procs=False, skip_existing=True, binary=True):
    ''' convert pysph output .npz files into vtk format

    Parameters
    ----------
    path : str
        directory where .npz files are located
    merge_procs : bool
        whether to merge the data from different procs into a single file
        (not yet implemented)
    skip_existing : bool
        skip files where corresponding vtk already exist
        this is useful if you've converted vtk files while a solver is running
        only want to convert the newly added files
    binary : bool
        whether to use binary format in vtk file
    The output vtk files are stored in a directory `solver_name` _vtk within
    the `path` directory

    '''
    if binary:
        data_mode = 'binary'
    else:
        data_mode = 'ascii'

    if merge_procs is True:
        # FIXME: implement
        raise NotImplementedError('merge_procs=True not implemented yet')

    solvers = get_output_details(path)
    for solver, (procs, entities, times) in solvers.items():
        print('converting solver:', solver)
        dir = os.path.join(path,solver+'_vtk')
        if not os.path.exists(dir):
            os.mkdir(dir)
        procs = sorted(procs)
        entities = sorted(entities)
        times = sorted(times, key=float)
        times_file = open(os.path.join(dir,'times'), 'w')

        for entity in entities:
            print('    entity:', entity)
            for proc in procs:
                print('        proc:', proc)
                print('        timesteps:', len(times))
                f = '%s_%s_%s_'%(solver,proc,entity)
                of = os.path.join(dir,f)

                for i, time in enumerate(times):
                    print('\r',i,)
                    if skip_existing and os.path.exists(f+str(i)):
                        continue
                    d = load(os.path.join(path, f+time+'.npz'))
                    arrs = {}
                    for nam,val in d.items():
                        if val.ndim > 0:
                            arrs[nam] = val
                    d.close()

                    scalars, vectors, tensors = detect_vectors_tensors(arrs)
                    vectors['V'] = ['u','v','w']
                    z = zeros_like(arrs['x'])
                    if 'v' not in arrs:
                        arrs['v'] = z
                    if 'w' not in arrs:
                        arrs['w'] = z
                    write_vtk(arrs, of+str(i),
                              scalars=scalars, vectors=vectors, tensors=tensors,
                              data_mode=data_mode)
                    times_file.write('%d\t%s\n'%(i,time))

        times_file.close()

def extract_text(path, particle_idx, props=['x','y','u','v','p','rho','sigma00','sigma01','sigma11'], ent=None, solvers=None):
    if solvers:
        raise NotImplementedError
    else:
        solvers = get_output_details(path)

    for solver, (procs, entities, times) in solvers.items():
        print('converting solver:', solver)
        dir = os.path.join(path,solver+'_vtk')
        if not os.path.exists(dir):
            os.mkdir(dir)
        procs = sorted(procs)
        entities = sorted(entities)
        times = sorted(times, key=float)
        times_file = open(os.path.join(dir,'times'), 'w')
        e = ent
        if ent is None:
            e = entities
        for entity in entities:
            if entity not in e:
                continue
            print('    entity:', entity)
            for proc in procs:
                print('        proc:', proc)
                print('        timesteps:', len(times))
                f = '%s_%s_%s_'%(solver,proc,entity)
                of = os.path.join(dir,f)
                files = [open(os.path.join(path,f+'%d.dat'%particle_id), 'w') for particle_id in particle_idx]
                print(files)
                for file in files:
                    file.write('i\tt\t'+'\t'.join(props))
                for i, time in enumerate(times):
                    print('\r',i,)
                    d = load(os.path.join(path, f+time+'.npz'))
                    s = '\n%d\t%s'%(i,time)
                    for j,file in enumerate(files):
                        file.write(s)
                        for prop in props:
                            file.write('\t')
                            file.write(str(d[prop][particle_idx[j]]))

                    d.close()

                for file in files:
                    file.close()

def test():
    l = ['x'+str(i) for i in range(3)]
    l.append('a0')
    l.append('a1')

    for i in range(3):
        for j in range(3):
            if i == j:
                l.append('XX%d'%i)
            if i <= j:
                l.append('S%d%d'%(i,j))
            l.append('T%d%d'%(i,j))

    scalars, vectors, tensors = detect_vectors_tensors(l)
    assert set(scalars) == set(['a0','a1'])
    assert set(vectors) == set(['x','XX'])
    assert set(tensors) == set(['S','T'])


if __name__ == '__main__':
    import sys
    pysph_to_vtk(path=sys.argv[1])

